from typing import Optional, TYPE_CHECKING

from requests import Response

from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from requests_futures.sessions import FuturesSession

    from mstrio.connection import Connection


@ErrorHandler(err_msg='Error getting information for a set of recipients.')
def get_recipients(
    connection,
    search_term,
    search_pattern="CONTAINS_ANY_WORD",
    offset=0,
    limit=-1,
    enabled_status='ALL',
):
    """Get information for a set of recipients.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        search_term (string): The value that the search_pattern parameter is set
            to. For example, if the search search_pattern is “Begins with”,
            this parameter would be the value that the search results would
            begin with.
        search_pattern (string): Available values : NONE, CONTAINS_ANY_WORD,
            BEGINS_WITH, BEGINS_WITH_PHRASE, EXACTLY, CONTAINS, ENDS_WITH.
            Default value : CONTAINS_ANY_WORD
        offset (int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit (int): Maximum number of items returned for a single search
            request. Used to control paging behavior. Use -1 (default ) for no
            limit (subject to governing settings).
        enabled_status (string): Specifies whether to return only enabled users
            or all users and user groups that match the search criteria.
            Available values : ALL, ENABLED_ONLY. Default value : ALL.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.get(
        url=f'{connection.base_url}/api/collaboration/recipients',
        params={
            'searchTerm': search_term,
            'searchPattern': search_pattern,
            'offset': offset,
            'limit': limit,
            'enabledStatus': enabled_status,
        },
    )


@ErrorHandler(err_msg='Error getting information for a set of users')
def get_users_info(
    connection,
    name_begins,
    abbreviation_begins,
    offset=0,
    limit=-1,
    fields=None,
    error_msg=None,
):
    """Get information for a set of users.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        name_begins (string): Characters that the user name must begin with.
        abbreviation_begins (string): Characters that the user abbreviation must
            begin with.
        offset (int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit (int): Maximum number of items returned for a single search
            request. Used to control paging behavior. Use -1 (default ) for no
            limit (subject to governing settings).
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.get(
        url=f'{connection.base_url}/api/users/',
        params={
            'nameBegins': name_begins,
            'abbreviationBegins': abbreviation_begins,
            'offset': offset,
            'limit': limit,
            'fields': fields,
        },
        headers={'X-MSTR-ProjectID': None},
    )


def get_users_info_async(
    future_session: "FuturesSession",
    connection,
    name_begins,
    abbreviation_begins,
    offset=0,
    limit=-1,
    fields=None,
):
    """Get information for a set of users asynchronously.

    Args:
        future_session: Future Session object to call MicroStrategy REST
            Server asynchronously
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        name_begins (string): Characters that the user name must begin with.
        abbreviation_begins (string): Characters that the user abbreviation must
            begin with.
        offset (int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit (int): Maximum number of items returned for a single search
            request. Used to control paging behavior. Use -1 (default ) for no
            limit (subject to governing settings).
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        Complete Future object.
    """
    params = {
        'nameBegins': name_begins,
        'abbreviationBegins': abbreviation_begins,
        'offset': offset,
        'limit': limit,
        'fields': fields,
    }
    url = f'{connection.base_url}/api/users/'
    headers = {'X-MSTR-ProjectID': None}
    future = future_session.get(url=url, headers=headers, params=params)
    return future


@ErrorHandler(err_msg='Error creating a new user with username: {username}')
def create_user(connection, body, username, fields=None):
    """Create a new user. The response includes the user ID, which other
    endpoints use as a request parameter to specify the user to perform an
    action on.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        body: JSON formatted user data;
                {
                "username": "string",
                "fullName": "string",
                "description": "string",
                "password": "string",
                "enabled": true,
                "passwordModifiable": true,
                "passwordExpirationDate": "2020-06-15T13:26:09.616Z",
                "requireNewPassword": true,
                "standardAuth": true,
                "ldapdn": "string",
                "trustId": "string",
                "memberships": [
                    "string"
                ]
                }
        username (str): username of a user, used in error message
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    return connection.post(
        url=f'{connection.base_url}/api/users',
        params={'fields': fields},
        json=body,
    )


@ErrorHandler(err_msg='Error getting addresses for user with ID {id}')
def get_addresses(connection, id, fields=None):
    """Get all of the addresses for a specific user.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string): User ID.
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    return connection.get(
        url=f'{connection.base_url}/api/users/{id}/addresses', params={'fields': fields}
    )


@ErrorHandler(err_msg='Error creating a new address for user with ID {id}')
def create_address(connection, id, body, fields=None):
    """Create a new address for a specific user.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string): User ID.
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        body: JSON-formatted address:
                {
                "name": "string",
                "deliveryMode": "EMAIL",
                "device": "GENERIC_EMAIL",
                "value": "string",
                "isDefault": true,
                "default": true
                }

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    return connection.post(
        url=f'{connection.base_url}/api/users/{id}/addresses',
        params={'fields': fields},
        json=body,
    )


@ErrorHandler(err_msg='Error creating a new address for user with ID {id}')
def create_address_v2(
    connection: "Connection", id: str, body: dict, fields: Optional[str] = None
) -> Response:
    """Create a new address for a specific user.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string): User ID.
        body (JSON): JSON-formatted address:
                {
                "name": "string",
                "physicalAddress": "string",
                "deliveryType": "string",
                "deviceId": "string",
                "deviceName": "string",
                "isDefault": boolean
                }
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.post(
        url=f'{connection.base_url}/api/v2/users/{id}/addresses',
        params={'fields': fields},
        json=body,
    )


@ErrorHandler(
    err_msg='Error updating address with ID {address_id} for user with ID {id}'
)
def update_address(connection, id, address_id, body, fields=None):
    """Update a specific address for a specific user.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string): User ID.
        address_id (string): Address ID.
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    return connection.put(
        url=f'{connection.base_url}/api/users/{id}/addresses/{address_id}',
        params={'fields': fields},
        json=body,
    )


@ErrorHandler(
    err_msg='Error deleting address with ID {address_id} for a user with ID {id}'
)
def delete_address(connection, id, address_id, fields=None):
    """Delete a specific address for a specific user.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string): User ID.
        address_id (string): Address ID.
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    return connection.delete(
        url=f'{connection.base_url}/api/users/{id}/addresses/{address_id}',
        headers={'X-MSTR-ProjectID': None},
        params={'fields': fields},
    )


@ErrorHandler(err_msg='Error getting security roles for a user with ID {id}')
def get_user_security_roles(connection, id, project_id=None):
    """Get all of the security roles for a specific user in a specific project.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): User ID

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.get(
        url=f'{connection.base_url}/api/users/{id}/securityRoles',
        params={'projectId': project_id},
    )


@ErrorHandler(err_msg="Error getting user {id} privileges for a project")
def get_user_privileges(connection, id, project_id=None, privilege_level=None):
    """Get user's privileges of a project including the source of the
    privileges.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): User ID
        project_id (string, optional): Project ID
        privilege_level (string, optional): Project Level Privilege

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    url = f'{connection.base_url}/api/users/{id}/privileges/'
    return connection.get(
        url=url,
        params={'privilege.level': privilege_level, 'projectId': project_id},
    )


@ErrorHandler(err_msg='Error getting user data usage limit for project with ID {id}')
def get_user_data_usage_limit(connection, id, project_id):
    """Get the data usage limit for users, either all users or a specific user,
    in a specific project. A typical use case would be that an administrator
    has set a project-level limit of data, such as 10GB, for all users in a
    specific project and has also limited the data usage for specific users,
    for example to 5GB.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): User ID
        project_id (str): Project ID

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    url = f'{connection.base_url}/api/users/{id}/projects/{project_id}/quotas'
    return connection.get(url=url)


@ErrorHandler(err_msg='Error getting information for a user with ID {id}')
def get_user_info(connection, id, fields=None):
    """Get information for a specific user.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string): User ID.
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    return connection.get(
        url=f'{connection.base_url}/api/users/{id}', params={'fields': fields}
    )


@ErrorHandler(err_msg='Error deleting user with ID {id}')
def delete_user(connection, id):
    """Delete user for specific user id.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string): User ID.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    return connection.delete(url=f'{connection.base_url}/api/users/{id}')


@ErrorHandler(err_msg='Error updating information for a user with ID: {id}')
def update_user_info(connection, id, body, fields=None):
    """Update specific information for a specific user.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string): User ID.
        body (JSON):
                    Body:{
                        "operationList": [
                            {
                            "op": "replace",
                            "path": "/username",
                            "value": "john1991"
                            }
                        ]
                        }
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    return connection.patch(
        url=f'{connection.base_url}/api/users/{id}',
        params={'fields': fields},
        json=body,
    )


@ErrorHandler(
    err_msg='Error getting information for the direct user groups that '
    'a user with ID {id} belongs to.'
)
def get_memberships(connection, id, fields=None):
    """Get information for the direct user groups that a specific user belongs
    to.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string): User ID.
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """

    return connection.get(
        url=f'{connection.base_url}/api/users/{id}/memberships',
        params={'fields': fields},
    )


@ErrorHandler(err_msg='Error getting security filters for user with ID {id}.')
def get_security_filters(
    connection: "Connection",
    id: str,
    projects: Optional[str | list[str]] = None,
    offset: int = 0,
    limit: int = -1,
    error_msg: Optional[str] = None,
):
    """Get each project level security filter and its corresponding inherited
    security filters for the user with given ID.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): User ID
        projects (str or list of str, optional): collection of projects' ids
            which is used for filtering data
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is -1.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/users/{id}/securityFilters"

    if projects and isinstance(projects, list):
        projects = ','.join(projects)

    params = {'projects.id': projects, 'offset': offset, 'limit': limit}
    return connection.get(url=url, params=params)



"""This is the demo script to show how to manage content cache. Its basic goal
is to present what can be done with this module and to ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.project_objects.content_cache import ContentCache

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

conn = get_connection(workstationData, PROJECT_NAME)

# Content cache management
# Listing caches with different conditions
list_of_all_caches = ContentCache.list_caches(connection=conn)

# Define a variable which can be later used in a script
CACHE_ID = $cache_id  # Insert ID of cache on which you want to perform actions

list_of_one_cache = ContentCache.list_caches(connection=conn, id=CACHE_ID)
list_of_caches_as_dicts = ContentCache.list_caches(connection=conn, to_dictionary=True)

# Get single content cache by its id
content_cache = ContentCache(connection=conn, cache_id=CACHE_ID)

# Unload content cache
content_cache.unload()

# Load content cache
content_cache.load()

# Refresh content cache instance data
content_cache.fetch()

# Listing properties of content cache
properties = content_cache.list_properties()

# Delete content cache
content_cache.delete(force=True)

# Define a variable which can be later used in a script
OTHER_CACHE_ID = $other_cache_id  # Insert ID of cache on which you want to perform actions

# Unload multiple content caches
ContentCache.unload_caches(connection=conn, cache_ids=[CACHE_ID, OTHER_CACHE_ID])

# Load multiple content caches
ContentCache.load_caches(connection=conn, cache_ids=[CACHE_ID, OTHER_CACHE_ID])

# Delete multiple content caches
ContentCache.delete_caches(connection=conn, cache_ids=[CACHE_ID, OTHER_CACHE_ID])



from typing import Callable, Optional

from mstrio.api import monitors
from mstrio.server.node import Node
from mstrio.utils.helper import _prepare_objects, auto_match_args, response_handler
from mstrio.utils.sessions import FuturesSessionWithRenewal


def all_nodes_async(
    connection,
    async_api: Callable,
    filters: dict,
    error_msg: str,
    unpack_value: Optional[str] = None,
    limit: Optional[int] = None,
    **kwargs,
):
    """Return list of objects fetched async using wrappers in monitors.py"""

    node = kwargs.get('node_name')
    if not node:
        nodes_response = monitors.get_node_info(connection).json()
        all_nodes = nodes_response['nodes']
        node_names = [node['name'] for node in all_nodes if node['status'] == 'running']
    else:
        node = node.name if isinstance(node, Node) else node
        if isinstance(node, str):
            node_names = [node]

    if kwargs.get('error_msg'):
        error_msg = kwargs.get('error_msg')

    with FuturesSessionWithRenewal(connection=connection, max_workers=8) as session:
        # Extract parameters of the api wrapper and set them using kwargs
        param_value_dict = auto_match_args(
            async_api,
            kwargs,
            exclude=[
                'connection',
                'limit',
                'offset',
                'future_session',
                'error_msg',
                'node_name',
            ],
        )
        futures = [
            async_api(
                future_session=session,
                connection=connection,
                node_name=n,
                **param_value_dict,
            )
            for n in node_names
        ]
        objects = []
        for f in futures:
            response = f.result()
            if not response.ok:
                response_handler(response, error_msg, throw_error=False)
            else:
                obj = _prepare_objects(response.json(), filters, unpack_value)
                objects.extend(obj)
    if limit:
        objects = objects[:limit]
    return objects



import logging
from typing import Optional, TYPE_CHECKING

from mstrio.api import objects
from mstrio.object_management.search_operations import (
    full_search,
    SearchDomain,
    SearchPattern,
)
from mstrio.types import ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.acl import ACLMixin
from mstrio.utils.certified_info import CertifiedInfo
from mstrio.utils.entity import CertifyMixin, CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.enum_helper import get_enum
from mstrio.utils.helper import (
    get_args_from_func,
    get_default_args_from_func,
    get_valid_project_id,
)

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


def list_objects(
    connection: "Connection",
    object_type: ObjectTypes | int,
    name: Optional[str] = None,
    project_id: Optional[str] = None,
    project_name: Optional[str] = None,
    domain: SearchDomain | int = SearchDomain.CONFIGURATION,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    to_dictionary: bool = False,
    limit: int = None,
    **filters,
) -> list["Object"] | list[dict]:
    """Get list of objects or dicts. Optionally filter the
    objects by specifying filters.

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        object_type: Object type. Possible values can
            be found in EnumDSSXMLObjectTypes
        name (string, optional): value the search pattern is set to, which
            will be applied to the names of objects being searched
        project_id: ID of the project for which the objects are to be listed
        project_name: project name
        domain: domain where the search will be performed,
            such as Local or Project, possible values
            are defined in EnumDSSXMLSearchDomain
        search_pattern (SearchPattern enum or int, optional): pattern to search
            for, such as Begin With or Exactly. Possible values are available in
            ENUM mstrio.object_management.SearchPattern.
            Default value is CONTAINS (4).
        to_dictionary: If True returns dict, by default (False) returns Objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg', 'owner', 'ext_type']

    Examples:
        >>> list_objects(connection, object_type=ObjectTypes.USER)
    """

    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=True,
    )

    result = Object._list_objects(
        connection=connection,
        object_type=object_type,
        name=name,
        project_id=project_id,
        domain=domain,
        pattern=search_pattern,
        to_dictionary=to_dictionary,
        limit=limit,
        **filters,
    )

    if not result:
        logger.info(
            f"Search in the domain: {domain.name} returned empty result."
            " Try searching in different domain."
        )
    return result


class Object(Entity, ACLMixin, CertifyMixin, CopyMixin, MoveMixin, DeleteMixin):
    """Class representing a general type object using attributes common for
     all available objects.

    Attributes:
        name: name of the object
        id: object ID
        description: Description of the object
        abbreviation: Object abbreviation
        version: Object version ID
        ancestors: List of ancestor folders
        type: Object type. Enum
        subtype: Object subtype
        ext_type: Object extended type.
        date_created: Creation time, "yyyy-MM-dd HH:mm:ss" in UTC
        date_modified: Last modification time, "yyyy-MM-dd HH:mm:ss" in UTC
        owner: User object that is the owner
        icon_path: Object icon path
        view_media: View media settings
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
        hidden: Specifies whether the object is hidden
        certified_info: Certification status, time of certification, and
            information about the certifier (currently only for document and
            report)
        project_id: project ID
        target_info: target information, only applicable to Shortcut objects
    """

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'certified_info': CertifiedInfo.from_dict,
        'owner': User.from_dict,
    }
    _API_GETTERS = {
        (
            'id',
            'name',
            'description',
            'date_created',
            'date_modified',
            'acg',
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acl',
            'comments',
            'project_id',
            'hidden',
            'target_info',
        ): objects.get_object_info
    }
    _API_PATCH: dict = {
        **Entity._API_PATCH,
        ('folder_id'): (objects.update_object, 'partial_put'),
    }

    def __init__(self, connection: "Connection", type: ObjectTypes, id: str):
        """Initialize object by ID.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str): Identifier of an existing object.
            type (ObjectTypes): object type
        """

        super().__init__(connection=connection, object_id=id, type=type)

    def _init_variables(self, **kwargs) -> None:
        self._OBJECT_TYPE = (
            ObjectTypes(kwargs.get("type"))
            if ObjectTypes.contains(kwargs.get("type"))
            else ObjectTypes.NONE
        )
        super()._init_variables(**kwargs)

    def __str__(self):
        return f"Object named: '{self.name}' with ID: '{self.id}'"

    def alter(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        abbreviation: Optional[str] = None,
    ) -> None:
        """Alter the object properties.

        Args:
            name: object name
            description: object description
            abbreviation: abbreviation
        """
        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    @classmethod
    def _list_objects(
        cls,
        connection: "Connection",
        object_type: ObjectTypes | int,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        to_dictionary: bool = False,
        domain: int | SearchDomain = SearchDomain.CONFIGURATION,
        pattern: int | SearchPattern = SearchPattern.CONTAINS,
        limit: Optional[int] = None,
        **filters,
    ) -> list["Object"] | list[dict]:
        objects = full_search(
            connection,
            object_types=get_enum(object_type, ObjectTypes),
            name=name,
            project=project_id,
            domain=domain,
            pattern=pattern,
            limit=limit,
            **filters,
        )
        if to_dictionary:
            return objects
        return [cls.from_dict(source=obj, connection=connection) for obj in objects]



import logging
from enum import auto
from functools import partial
from typing import Optional, Type

import pandas as pd

from mstrio.api import incremental_refresh_reports as refresh_api
from mstrio.api import objects
from mstrio.connection import Connection
from mstrio.modeling import (
    Expression,
    ExpressionFormat,
    ObjectSubType,
    SchemaObjectReference,
)
from mstrio.object_management import Folder, SearchPattern, full_search
from mstrio.project_objects import OlapCube
from mstrio.project_objects.incremental_refresh_report import (
    AdvancedProperties,
    Template,
)
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import (
    delete_none_values,
    filter_params_for_func,
    get_objects_id,
    get_valid_project_id,
)
from mstrio.utils.parser import Parser
from mstrio.utils.version_helper import class_version_handler, method_version_handler
from mstrio.utils.wip import WipLevels, module_wip

logger = logging.getLogger(__name__)

module_wip(globals(), level=WipLevels.WARNING)


@method_version_handler('11.3.0600')
def list_incremental_refresh_reports(
    connection: Connection,
    name: Optional[str] = None,
    pattern: SearchPattern | int = SearchPattern.CONTAINS,
    project_id: Optional[str] = None,
    project_name: Optional[str] = None,
    to_dictionary: bool = False,
    limit: Optional[int] = None,
    folder_id: Optional[str] = None,
    filters: Optional[dict] = None,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    show_filter_tokens: bool = False,
    show_advanced_properties: bool = False,
) -> list[Type['IncrementalRefreshReport']] | list[dict]:
    return IncrementalRefreshReport.list(
        connection,
        name,
        pattern,
        project_id,
        project_name,
        to_dictionary,
        limit,
        folder_id,
        filters,
        show_expression_as,
        show_filter_tokens,
        show_advanced_properties,
    )


@class_version_handler('11.3.0600')
class IncrementalRefreshReport(Entity, CopyMixin, MoveMixin, DeleteMixin):
    class IncrementType(AutoName):
        REPORT = auto()
        FILTER = auto()

    class RefreshType(AutoName):
        INSERT = auto()
        UPDATE = auto()
        UPDATE_ONLY = auto()
        REPLACE = auto()
        DELETE = auto()

    _OBJECT_TYPE = ObjectTypes.REPORT_DEFINITION

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'target_cube': SchemaObjectReference.from_dict,
        'increment_type': IncrementType,
        'refresh_type': RefreshType,
        'template': Template.from_dict,
        'filter': Expression.from_dict,
        'advanced_properties': AdvancedProperties.from_dict,
    }

    _API_GETTERS = {
        (
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'ancestors',
            'acg',
            'acl',
        ): objects.get_object_info,
        (
            'id',
            'name',
            'description',
            'sub_type',
            'date_created',
            'date_modified',
            'path',
            'version_id',
            'is_embedded',
            'primary_locale',
            'destination_folder_id',
            'target_cube',
            'refresh_type',
            'increment_type',
            'template',
            'filter',
            'advanced_properties',
        ): refresh_api.get_incremental_refresh_report,
    }
    _API_PATCH: dict = {
        (
            'target_cube',
            'refresh_type',
            'increment_type',
            'template',
            'filter',
            'advanced_properties',
        ): (refresh_api.update_incremental_refresh_report, "put"),
        (
            'name',
            'description',
            'destination_folder_id',
            'is_embedded',
            'folder_id',
        ): (objects.update_object, 'partial_put'),
    }
    _PATCH_PATH_TYPES = {
        'name': str,
        'description': str,
        'destination_folder_id': str,
        'is_embedded': bool,
        'target_cube': dict,
        'refresh_type': str,
        'increment_type': str,
        'filter': dict,
        'advanced_properties': dict,
    }

    _API_DELETE = staticmethod(objects.delete_object)
    _ALLOW_NONE_ATTRIBUTES = ['filter', 'template']
    _KEEP_CAMEL_CASE = ['vldbProperties', 'metricJoinTypes', 'attributeJoinTypes']

    def __init__(
        self,
        connection: Connection,
        id: Optional[str] = None,
        name: Optional[str] = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        show_filter_tokens: bool = False,
        show_advanced_properties: bool = False,
    ):
        connection._validate_project_selected()

        if id is None:
            reports = super()._find_object_with_name(
                connection=connection,
                name=name,
                listing_function=partial(self.list, pattern=SearchPattern.EXACTLY),
            )
            id = reports['id']

        super().__init__(
            connection=connection,
            object_id=id,
            show_expression_as=show_expression_as,
            show_filter_tokens=show_filter_tokens,
            show_advanced_properties=show_advanced_properties,
        )

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)

        self.primary_locale = kwargs.get('primary_locale')
        self.is_embedded = kwargs.get('is_embedded')
        self.destination_folder_id = kwargs.get('destination_folder_id')

        self._sub_type = (
            ObjectSubType(kwargs.get('sub_type')) if kwargs.get('sub_type') else None
        )
        self._path = kwargs.get('path')

        self.target_cube = (
            SchemaObjectReference.from_dict(kwargs.get('target_cube'))
            if kwargs.get('target_cube')
            else None
        )
        self.refresh_type = (
            self.RefreshType(kwargs.get('refresh_type'))
            if kwargs.get('refresh_type')
            else None
        )
        self.increment_type = (
            self.IncrementType(kwargs.get('increment_type'))
            if kwargs.get('increment_type')
            else None
        )
        self.template = (
            Template.from_dict(kwargs.get('template'), self.connection)
            if kwargs.get('template')
            else None
        )
        self.filter = (
            Expression.from_dict(kwargs.get('filter'), self.connection)
            if kwargs.get('filter')
            else None
        )
        self.advanced_properties = (
            AdvancedProperties.from_dict(
                kwargs.get('advanced_properties'), self.connection
            )
            if kwargs.get('advanced_properties')
            else None
        )

        show_expression_as = kwargs.get('show_expression_as', 'tree')
        self._show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )
        self._show_filter_tokens = kwargs.get('show_filter_tokens', False)
        self._show_advanced_properties = kwargs.get('show_advanced_properties', False)

    def execute(
        self,
        fields: Optional[str] = None,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ) -> None:
        project_id = get_valid_project_id(
            self.connection, project_id, project_name, with_fallback=True
        )

        response = refresh_api.execute_incremental_refresh_report(
            self.connection,
            id=self.id,
            project_id=project_id,
            fields=fields,
        )

        if response.ok:
            logger.info(
                f"Execution of Incremental Refresh Report: '{self.name}' has been "
                f"successfully scheduled under job: {response.json()}."
            )

    @classmethod
    def list(
        cls,
        connection: Connection,
        name: Optional[str] = None,
        pattern: SearchPattern | int = SearchPattern.CONTAINS,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        to_dictionary: bool = False,
        limit: Optional[int] = None,
        folder_id: Optional[str] = None,
        filters: Optional[dict] = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        show_filter_tokens: bool = False,
        show_advanced_properties: bool = False,
    ) -> list[Type['IncrementalRefreshReport']] | list[dict]:
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=False if project_name else True,
        )
        filters = filters or {}

        if folder_id:
            filters = filters | {'root': folder_id}

        objects = full_search(
            connection,
            object_types=ObjectSubTypes.INCREMENTAL_REFRESH_REPORT,
            project=project_id,
            name=name,
            pattern=pattern,
            limit=limit,
            **filters,
        )

        if to_dictionary:
            return objects

        return [
            cls.from_dict(
                obj
                | {
                    'show_expression_as': show_expression_as
                    if isinstance(show_expression_as, ExpressionFormat)
                    else ExpressionFormat(show_expression_as),
                    'show_filter_tokens': show_filter_tokens,
                    'show_advanced_properties': show_advanced_properties,
                },
                connection,
            )
            for obj in objects
        ]

    @classmethod
    def create(
        cls,  # NOSONAR
        connection: Connection,
        name: str,
        destination_folder: Folder | str,
        target_cube: OlapCube | SchemaObjectReference | dict,
        increment_type: IncrementType | str,
        refresh_type: RefreshType | str,
        template: Template | dict = None,
        filter: Optional[Expression | dict] = None,
        advanced_properties: Optional[dict] = None,
        description: Optional[str] = None,
        primary_locale: Optional[str] = None,
        is_embedded: bool = False,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        show_filter_tokens: bool = False,
        show_advanced_properties: bool = False,
    ) -> 'IncrementalRefreshReport':
        information = {
            'name': name,
            'description': description,
            'destinationFolderId': get_objects_id(destination_folder, Folder),
            'primaryLocale': primary_locale,
            'isEmbedded': is_embedded,
        }

        if isinstance(target_cube, SchemaObjectReference):
            target_cube = target_cube.to_dict()
        elif isinstance(target_cube, OlapCube):
            target_cube = SchemaObjectReference(
                object_id=target_cube.id,
                name=target_cube.name,
                sub_type=ObjectSubType.REPORT_CUBE,
            ).to_dict()

        body = {
            'information': delete_none_values(information, recursion=True),
            'targetCube': target_cube,
            'incrementType': get_enum_val(increment_type, cls.IncrementType),
            'refreshType': get_enum_val(refresh_type, cls.RefreshType),
            'template': template.to_dict()
            if isinstance(template, Template)
            else template,
            'filter': filter.to_dict()
            if isinstance(filter, Expression)
            else (filter or {}),
            'advanced_properties': advanced_properties,
        }

        response = refresh_api.create_incremental_refresh_report(
            connection=connection,
            body=body,
            show_expression_as=get_enum_val(show_expression_as, ExpressionFormat),
            show_filter_tokens=show_filter_tokens,
            show_advanced_properties=show_advanced_properties,
        ).json()

        logger.info(
            f"Successfully created Incremental Refresh Report named: "
            f"'{name}' with ID: '{response['id']}'"
        )

        return cls.from_dict(
            source={
                **response,
                'show_expression_as': show_expression_as,
                'show_filter_tokens': show_filter_tokens,
                'show_advanced_properties': show_advanced_properties,
            },
            connection=connection,
        )

    @classmethod
    def create_from_cube(
        cls,
        connection: Connection,
        name: str,
        destination_folder: Folder | str,
        target_cube: OlapCube | SchemaObjectReference | dict,
        refresh_type: RefreshType | str,
        filter: Optional[Expression | dict] = None,
        advanced_properties: Optional[dict] = None,
        description: Optional[str] = None,
        primary_locale: Optional[str] = None,
        is_embedded: bool = False,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        show_filter_tokens: bool = False,
        show_advanced_properties: bool = False,
    ) -> 'IncrementalRefreshReport':
        parameters = filter_params_for_func(cls.create_from_cube, locals(), ['cls'])

        return cls.create(**parameters, increment_type=cls.IncrementType.REPORT)

    def alter(
        self,
        name: Optional[str] = None,
        target_cube: Optional[OlapCube | SchemaObjectReference | dict] = None,
        increment_type: Optional[IncrementType | str] = None,
        refresh_type: Optional[RefreshType | str] = None,
        template: Optional[Template | dict] = None,
        filter: Optional[Expression | dict] = None,
        advanced_properties: Optional[dict] = None,
        description: Optional[str] = None,
        primary_locale: Optional[str] = None,
        is_embedded: Optional[bool] = None,
    ):
        if isinstance(target_cube, OlapCube):
            target_cube = SchemaObjectReference(
                object_id=target_cube.id,
                name=target_cube.name,
                sub_type=ObjectSubType.REPORT_CUBE,
            )
        filter = {} if self.filter is None and filter is None else filter  # NOSONAR
        arguments = filter_params_for_func(self.alter, locals(), exclude=['self'])

        self._alter_properties(**arguments)

    def change_increment_type_to_filter(self) -> None:
        self.fetch()
        self.template = None
        self.alter(increment_type=IncrementalRefreshReport.IncrementType.FILTER)

    def reset_template_to_default(self) -> None:
        if self.increment_type == IncrementalRefreshReport.IncrementType.REPORT:
            self.change_increment_type_to_filter()
            self.alter(increment_type=IncrementalRefreshReport.IncrementType.REPORT)

    def list_applicable_vldb_properties(self) -> dict:
        response = refresh_api.get_incremental_refresh_report_vldb_properties(
            self.connection, id=self.id, project_id=self.connection.project_id
        )
        return response.json()

    def get_preview_data(
        self,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        fields: Optional[str] = None,
    ) -> pd.DataFrame:
        response = refresh_api.create_incremental_refresh_report_instance(
            self.connection, self.id
        )
        instance_id = response.headers['X-MSTR-MS-Instance']
        refresh_api.request_incremental_refresh_report_preview_data(
            connection=self.connection,
            id=self.id,
            instance_id=instance_id,
            project_id=self.connection.project_id,
            fields=fields,
        )
        json = refresh_api.get_incremental_refresh_report_preview_data(
            connection=self.connection,
            id=self.id,
            instance_id=instance_id,
            project_id=self.connection.project_id,
            offset=offset,
            limit=limit,
            fields=fields,
        ).json()

        refresh_api.delete_incremental_refresh_report_instance(
            connection=self.connection,
            id=self.id,
            instance_id=instance_id,
        )

        parser = Parser(response=json, parse_cube=False)
        parser.parse(response=json)

        return parser.dataframe



from enum import auto, Enum

from mstrio.utils.enum_helper import AutoName


class Function(AutoName):
    """Enumeration constant used to specify the function used for calculation
    in expression nodes"""

    THIRD_PARTY = auto()
    CUSTOM = auto()
    PLUS = auto()
    MINUS = auto()
    TIMES = auto()
    DIVIDE = auto()
    UNARY_MINUS = auto()
    EQUALS = auto()
    NOT_EQUAL = auto()
    GREATER = auto()
    LESS = auto()
    GREATER_EQUAL = auto()
    LESS_EQUAL = auto()
    SUM = auto()
    COUNT = auto()
    AVG = auto()
    MIN = auto()
    MAX = auto()
    BETWEEN = auto()
    LIKE = auto()
    AND = auto()
    OR = auto()
    NOT = auto()
    IN = auto()
    RANK = auto()
    ABS = auto()
    RUNNING_SUM = auto()
    RUNNING_AVG = auto()
    MOVING_SUM = auto()
    MOVING_AVG = auto()
    PRODUCT = auto()
    MEDIAN = auto()
    MODE = auto()
    STDEV = auto()
    VAR = auto()
    GEOMEAN = auto()
    EQUAL_ENHANCED = auto()
    NOT_EQUAL_ENHANCED = auto()
    GREATER_EQUAL_ENHANCED = auto()
    LESS_EQUAL_ENHANCED = auto()
    BETWEEN_ENHANCED = auto()
    BANDING = auto()
    BANDING_C = auto()
    BANDING_P = auto()
    NOT_LIKE = auto()
    NOT_BETWEEN = auto()
    INTERSECT = auto()
    INTERSECT_IN = auto()
    NULL_TO_ZERO = auto()
    ZERO_TO_NULL = auto()
    APPLY_SIMPLE = auto()
    APPLY_AGGREGATION = auto()
    APPLY_LOGIC = auto()
    APPLY_COMPARISON = auto()
    APPLY_RELATIVE = auto()
    IS_NULL = auto()
    IS_NOT_NULL = auto()
    UCASE = auto()
    NOT_IN = auto()
    N_TILE = auto()
    PERCENTILE = auto()
    MOVING_MAX = auto()
    MOVING_MIN = auto()
    MOVING_DIFFERENCE = auto()
    MOVING_STDEV = auto()
    EXP_WGH_MOVING_AVG = auto()
    MOVING_COUNT = auto()
    RUNNING_MAX = auto()
    RUNNING_MIN = auto()
    RUNNING_STDEV = auto()
    RUNNING_COUNT = auto()
    EXP_WGH_RUNNING_AVG = auto()
    NOT_BETWEEN_ENHANCED = auto()
    CONCAT = auto()
    FIRST_IN_RANGE = auto()
    LAST_IN_RANGE = auto()
    VALUE_SEGMENT = auto()
    CONTAINS = auto()
    BEGINS_WITH = auto()
    ENDS_WITH = auto()
    NOT_CONTAINS = auto()
    NOT_BEGINS_WITH = auto()
    NOT_ENDS_WITH = auto()
    CASE = auto()
    CASE_V = auto()
    STDEV_P = auto()
    RUNNING_STDEV_P = auto()
    MOVING_STDEV_P = auto()
    N_TILE_S = auto()
    N_TILE_VS = auto()
    VAR_P = auto()
    CURRENT_DATE = auto()
    DAY_OF_MONTH = auto()
    DAY_OF_WEEK = auto()
    DAY_OF_YEAR = auto()
    WEEK = auto()
    MONTH = auto()
    QUARTER = auto()
    YEAR = auto()
    CURRENT_DATE_TIME = auto()
    CURRENT_TIME = auto()
    HOUR = auto()
    MINUTE = auto()
    SECOND = auto()
    MILLI_SECOND = auto()
    CONCAT_NO_BLANK = auto()
    LENGTH = auto()
    LOWER = auto()
    L_TRIM = auto()
    POSITION = auto()
    R_TRIM = auto()
    SUB_STR = auto()
    INIT_CAP = auto()
    TRIM = auto()
    RIGHT_STR = auto()
    LEFT_STR = auto()
    GREATEST = auto()
    LEAST = auto()
    FIRST = auto()
    LAST = auto()
    DATE = auto()
    DAYS_BETWEEN = auto()
    MONTHS_BETWEEN = auto()
    ADD_DAYS = auto()
    ADD_MONTHS = auto()
    MONTH_START_DATE = auto()
    MONTH_END_DATE = auto()
    YEAR_START_DATE = auto()
    YEAR_END_DATE = auto()
    IF = auto()
    APPLY_OPTIONAL = auto()
    APPLY_CS_SECURITY_FILTER = auto()
    UNION = auto()
    EXCEPT = auto()
    COALESCE = auto()
    ADD = auto()
    AVERAGE = auto()
    MULTIPLY = auto()
    BANDING_M = auto()
    OLAP_SUM = auto()
    OLAP_AVG = auto()
    OLAP_COUNT = auto()
    OLAP_MAX = auto()
    OLAP_MIN = auto()
    LAG = auto()
    LEAD = auto()
    OLAP_RANK = auto()
    REPEAT = auto()
    BIT_AND = auto()
    BIT_OR = auto()
    BIT_XOR = auto()
    BIT_NOT = auto()
    BIT_LEFT_SHIFT = auto()
    BIT_RIGHT_SHIFT = auto()
    AMPERSAND = auto()
    ORDINAL_RANK = auto()
    HISTOGRAM_MEDIAN = auto()
    BAND_NAMES = auto()
    PERCENT_RANK_RELATIVE = auto()
    SEARCH = auto()
    IF_BY_DIMTY = auto()
    GET_EXTRA_OUTPUT = auto()
    TO_DATE_TIME = auto()
    QUARTER_START_DATE = auto()
    WEEK_START_DATE = auto()
    TO_STRING = auto()
    TO_NUMBER = auto()
    PERCENT_RANK = auto()
    STR_REPLACE = auto()
    STR_MATCH = auto()
    STR_SPLIT = auto()
    STR_CHAR = auto()
    STR_REPEAT = auto()
    DATE_DIFF = auto()
    STR_BEGINS_WITH = auto()
    STR_ENDS_WITH = auto()
    WEIGHT_STD_P = auto()
    WEIGHT_MEAN_AVE = auto()
    CONCAT_AGG = auto()
    WEIGHT_COV = auto()
    WEIGHT_CORR = auto()
    STR_LAST_POSITION = auto()
    STR_TITLE_CAP = auto()
    FISCAL_WEEK = auto()
    FISCAL_MONTH = auto()
    FISCAL_QUARTER = auto()
    FISCAL_YEAR = auto()
    NULL_TO_EMPTY = auto()
    DESCENDANTS = auto()
    ANCESTORS = auto()
    PARENTS = auto()
    CHILDREN = auto()
    TUPLE = auto()


class ExpressionType(AutoName):
    """Enumeration constant indicating the expression type of expression node"""

    DYNAMIC = auto()
    STATIC = auto()
    GENERIC = auto()
    FILTER_SINGLE_BASE_FORM_QUAL = auto()
    FILTER_MULTI_BASE_FORM_QUAL = auto()
    FILTER_JOINT_FORM_QUAL = auto()
    FILTER_LIST_QUAL = auto()
    FILTER_LIST_FORM_QUAL = auto()
    FILTER_JOINT_LIST_QUAL = auto()
    FILTER_JOINT_LIST_FORM_QUAL = auto()
    FILTER_SINGLE_BASE_FORM_EXPRESSION = auto()
    FILTER_SINGLE_METRIC_QUAL = auto()
    FILTER_MULTI_METRIC_QUAL = auto()
    FILTER_METRIC_EXPRESSION = auto()
    FILTER_EMBED_QUAL = auto()
    FILTER_BRANCH_QUAL = auto()
    FILTER_RELATIONSHIP_QUAL = auto()
    FILTER_ALL_ATTRIBUTE_QUAL = auto()
    FILTER_ATTRIBUTE_ID_QUAL = auto()
    FILTER_ATTRIBUTE_DESC_QUAL = auto()
    AGG_METRIC = auto()
    BANDING = auto()
    FILTER_REPORT_QUAL = auto()
    MDX_SAP_VARIABLE = auto()
    SQL_QUERY_QUAL = auto()
    CANCELED_PROMPT = auto()
    ELEMENT_LIST = auto()
    ELEMENT_SINGLE = auto()
    FORM_OF_THIS_ATTRIBUTE = auto()
    CSI_UPDATE = auto()
    CSI_INSERT = auto()
    CSI_DELETE = auto()
    CSI_GROUP = auto()


class DimtyType(AutoName):
    """Enumeration constant  indicating the dimty type of expression node"""

    NONE = auto()
    CONTINUATION = auto()
    EXCLUSIVE_CONTINUATION = auto()
    OUTPUT_LEVEL = auto()
    BREAK_BY = auto()
    EMBEDDED = auto()
    UNSPECIFIED = auto()
    PARAMETER = auto()


class DependenceType(AutoName):
    """Enumeration constant indicating the dependence type of expression node"""

    DEFAULT = auto()
    INDEPENDENT = auto()
    DEPENDENT = auto()


class NodeType(AutoName):
    """Enumeration constant indicating the type of node within
    the expression tree
    """

    OPERATOR = auto()
    OBJECT_REFERENCE = auto()
    COLUMN_REFERENCE = auto()
    CONSTANT = auto()
    RELATIONSHIP = auto()
    FORM_SHORTCUT = auto()
    DYNAMIC_DATE_TIME = auto()
    PREDICATE_CUSTOM = auto()
    PREDICATE_PROMPT_QUALIFICATION = auto()
    PREDICATE_METRIC_QUALIFICATION = auto()
    PREDICATE_RELATIONSHIP = auto()
    PREDICATE_JOINT_ELEMENT_LIST = auto()
    PREDICATE_ELEMENT_LIST = auto()
    PREDICATE_FORM_QUALIFICATION = auto()
    PREDICATE_FILTER_QUALIFICATION = auto()
    PREDICATE_REPORT_QUALIFICATION = auto()
    PREDICATE_BANDING_SIZE = auto()
    PREDICATE_BANDING_COUNT = auto()
    PREDICATE_BANDING_POINTS = auto()
    PREDICATE_BANDING_DISTINCT = auto()


class IsIndependent(Enum):
    """Flag that indicates whether a node will be evaluated independently
    of other parts of the larger expression"""

    YES = 1
    NO = 0


class ExpressionFormat(AutoName):
    """ "Expression format to be fetched from server, it might be tree or token:
    - tree: tree data structure fully defining the expression. This format can
    be used if you want to examine and modify the expression programmatically.
    - tokens: list of parsed tokens. This format can be used if you want
    to examine and modify the expression using the parser component. Note that
    generating tokens requires additional time.
    """

    TREE = auto()
    TOKENS = auto()



"""This is the demo script to show how to manage transmitters.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.distribution_services import (
    EmailTransmitterProperties,
    list_transmitters,
    RecipientFieldType,
    Transmitter,
    TransmitterDeliveryType
)

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Define variables which can be later used in a script
SENDER_DISPLAY_NAME = $sender_display_name  # name of a sender
SENDER_EMAIL_ADDRESS = $sender_email_address  # address of a sender
REPLY_TO_DISPLAY_NAME = $reply_to_display_name  # name for destination for replies
REPLY_TO_EMAIL_ADDRESS = $reply_to_email_address  # address for destination for replies

# get list of transmitters
transmitters = list_transmitters(conn)

# Define variables which can be later used in a script
TRANSMITTER_ID = $transmitter_id
TRANSMITTER_NAME = $transmitter_name

# get transmitter by id
transmitter = Transmitter(conn, id=TRANSMITTER_ID)

# get transmitter by name
transmitter = Transmitter(conn, name=TRANSMITTER_NAME)

# list properties for transmitter
transmitter.list_properties()

# create email transmitter properties object
# which is used when creating transmitter
etp = EmailTransmitterProperties(
    sender_display_name=SENDER_DISPLAY_NAME,
    sender_email_address=SENDER_EMAIL_ADDRESS,
    reply_to_display_name=REPLY_TO_DISPLAY_NAME,
    reply_to_email_address=REPLY_TO_EMAIL_ADDRESS,
    # see distribution_services/transmitter/transmitter.py - RecipientFieldType
    # for available values
    recipient_field_type=RecipientFieldType.TO,
    save_message_to_file=False,
    send_message_to_file=False,
    send_message_via_smtp=False,
    save_file_path=None,
    notify_on_success=False,
    notify_on_failure=False,
    notification_email_address=None,
)

# Define variables which can be later used in a script
TRANSMITTER_DESCRIPTION = $transmitter_description
NOTIFICATION_EMAIL_ADDRESS = $notification_email_address

# create a transmitter with delivery type as `email` (when type is `email` then
# it is mandatory to provide `email_transmitter_properties`). Delivery type
# can be also passed as a string and `email_transmitter_properties` can be
# passed as a dictionary
new_t = Transmitter.create(
    connection=conn,
    name=TRANSMITTER_NAME,
    delivery_type=TransmitterDeliveryType.EMAIL,
    description=TRANSMITTER_DESCRIPTION,
    email_transmitter_properties=etp,
)

# alter email transmitter and its properties
etp.notify_on_success = True
etp.notify_on_failure = True
etp.notification_email_address = NOTIFICATION_EMAIL_ADDRESS
etp.save_message_to_file = True
new_t.alter(
    name=TRANSMITTER_NAME,
    description=TRANSMITTER_DESCRIPTION,
    email_transmitter_properties=etp,
)

# create file transmitter
new_file_t = Transmitter.create(
    connection=conn,
    name=TRANSMITTER_NAME,
    delivery_type=TransmitterDeliveryType.FILE,
    description=TRANSMITTER_DESCRIPTION,
)

# Define variables which can be later used in a script
TRANSMITTER_NEW_NAME = $transmitter_new_name
TRANSMITTER_NEW_DESCRIPTION = $transmitter_new_description

# alter file transmitter
new_file_t.alter(name=TRANSMITTER_NEW_NAME, description=TRANSMITTER_NEW_DESCRIPTION)

# create ipad transmitter
new_ipad_t = Transmitter.create(
    connection=conn,
    name=TRANSMITTER_NAME,
    delivery_type=TransmitterDeliveryType.IPAD,
    description=TRANSMITTER_DESCRIPTION,
)

# alter ipad transmitter
new_ipad_t.alter(name=TRANSMITTER_NEW_NAME, description=TRANSMITTER_NEW_DESCRIPTION)

# create iphone transmitter
new_iphone_t = Transmitter.create(
    connection=conn,
    name=TRANSMITTER_NAME,
    delivery_type=TransmitterDeliveryType.IPHONE,
    description=TRANSMITTER_DESCRIPTION,
)

# alter iphone transmitter
new_iphone_t.alter(name=TRANSMITTER_NEW_NAME, description=TRANSMITTER_NEW_DESCRIPTION)

# create print transmitter
new_print_t = Transmitter.create(
    connection=conn,
    name=TRANSMITTER_NAME,
    delivery_type=TransmitterDeliveryType.PRINT,
    description=TRANSMITTER_DESCRIPTION,
)

# alter print transmitter
new_print_t.alter(name=TRANSMITTER_NEW_NAME, description=TRANSMITTER_NEW_DESCRIPTION)

# Delete transmitter. When argument `force` is set to `False` (default value),
# then deletion must be confirmed by selecting appropriate prompt value.
transmitter.delete(force=True)



from typing import Optional, TYPE_CHECKING

from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.connection import Connection


@ErrorHandler(err_msg='Error listing Transmitters.')
def get_transmitters(connection: "Connection", error_msg: Optional[str] = None):
    """Get a list of all transmitters that the authenticated user has access
    to.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/transmitters"
    return connection.get(url=url)


@ErrorHandler(err_msg='Error creating Transmitter.')
def create_transmitter(connection: "Connection", body, error_msg: Optional[str] = None):
    """Create a new transmitter.

    Args:
        connection: MicroStrategy REST API connection object
        body: Transmitter creation body
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    url = f"{connection.base_url}/api/transmitters"
    return connection.post(url=url, json=body)


@ErrorHandler(err_msg='Error getting Transmitter with ID {id}')
def get_transmitter(connection: "Connection", id: str, error_msg: Optional[str] = None):
    """Get transmitter by a specific id.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the transmitter
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/transmitters/{id}"
    return connection.get(url=url)


@ErrorHandler(err_msg='Error updating Transmitter with ID {id}')
def update_transmitter(
    connection: "Connection", id: str, body: dict, error_msg: Optional[str] = None
):
    """Update a transmitter.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the transmitter
        body: Transmitter update info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/transmitters/{id}"
    return connection.put(url=url, json=body)


@ErrorHandler(err_msg='Error deleting Transmitter with ID {id}')
def delete_transmitter(
    connection: "Connection", id: str, error_msg: Optional[str] = None
):
    """Delete a transmitter.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the transmitter
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    url = f"{connection.base_url}/api/transmitters/{id}"
    return connection.delete(url=url)



"""Delete all subscription in all projects which owners are departed users.
For which project message about successful (or not) deletion will be printed.
"""

from mstrio.api.projects import get_projects
from mstrio.connection import Connection, get_connection
from mstrio.distribution_services import SubscriptionManager
from mstrio.users_and_groups import list_users


def delete_subscriptions_of_departed_users(connection: "Connection") -> None:
    """Delete all subscription in all projects which owners are departed users.

    Args:
        Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
    """

    # get all projects that the authenticated user has access to
    response = get_projects(connection, whitelist=[('ERR014', 403)])
    projects_ = response.json() if response.ok else []
    # get all disabled users
    all_users = list_users(connection=connection)
    disabled_users = [u for u in all_users if not u.enabled]

    for project_ in projects_:
        project_id = project_['id']
        sub_manager = SubscriptionManager(connection=connection, project_id=project_id)
        for user_ in disabled_users:
            subs = sub_manager.list_subscriptions(owner={'id': user_.id})
            msg = f"subscriptions of user with ID: {user_.id}"
            msg += f" in project {project_['name']} with ID: {project_['id']}"
            # call of the function below returns True if all passed
            # subscriptions were deleted
            if sub_manager.delete(subscriptions=subs, force=True):
                print("All " + msg + " were deleted.", flush=True)
            else:
                print(
                    "Not all " + msg + " were deleted or there was no subscriptions.",
                    flush=True,
                )


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# execute deletion of addresses from departed users
delete_subscriptions_of_departed_users(conn)



from typing import Optional, TYPE_CHECKING

from requests import Response

from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.helper import exception_handler, response_handler

if TYPE_CHECKING:
    from requests_futures.sessions import FuturesSession

    from mstrio.connection import Connection


@ErrorHandler(err_msg='Error getting subscription list.')
def list_subscriptions(
    connection, project_id, fields=None, offset=0, limit=-1, error_msg=None
):
    """Get a list of subscriptions.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        project_id (str): ID of the project
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.get(
        url=f'{connection.base_url}/api/subscriptions',
        params={'offset': offset, 'limit': limit, 'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
    )


def list_subscriptions_async(
    future_session: "FuturesSession",
    connection,
    project_id,
    fields=None,
    offset=0,
    limit=-1,
):
    """Get a list of subscriptions asynchronously.

    Args:
        future_session: Future Session object to call MicroStrategy REST
            Server asynchronously
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        project_id (str): ID of the project
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).

    Returns:
        Complete Future object.
    """
    params = {'offset': offset, 'limit': limit, 'fields': fields}
    url = f'{connection.base_url}/api/subscriptions'
    headers = {'X-MSTR-ProjectID': project_id}

    return future_session.get(url=url, headers=headers, params=params)


@ErrorHandler(err_msg='Error getting Dynamic Recipient List list.')
def list_dynamic_recipient_lists(
    connection: "Connection",
    project_id: str,
    offset: int = 0,
    limit: int = -1,
    fields: Optional[str] = None,
    error_msg: Optional[str] = None,
) -> Response:
    """Get a list of Dynamic Recipient Lists.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        project_id (str): ID of the project
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.get(
        url=f'{connection.base_url}/api/dynamicRecipientLists',
        params={'offset': offset, 'limit': limit, 'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
    )


def list_dynamic_recipient_lists_async(
    future_session: "FuturesSession",
    connection: "Connection",
    project_id: str,
    offset: int = 0,
    limit: int = -1,
    fields: Optional[str] = None,
) -> Response:
    """Get a list of Dynamic Recipient Lists asynchronously.

    Args:
        future_session: Future Session object to call MicroStrategy REST
            Server asynchronously
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        project_id (str): ID of the project
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        Complete Future object.
    """
    return future_session.get(
        url=f'{connection.base_url}/api/dynamicRecipientLists',
        headers={'X-MSTR-ProjectID': project_id},
        params={'offset': offset, 'limit': limit, 'fields': fields},
    )


@ErrorHandler(err_msg='Error getting subscription {subscription_id} information.')
def get_subscription(
    connection, subscription_id, project_id, fields=None, error_msg=None
):
    """Get information of a specific subscription for a given project.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        subscription_id (str): ID of the subscription
        project_id (str): ID of the project
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.get(
        url=f'{connection.base_url}/api/subscriptions/{subscription_id}',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg='Error getting Dynamic Recipient List {list_id} information.')
def get_dynamic_recipient_list(
    connection: "Connection",
    id: str,
    project_id: str,
    fields: Optional[str] = None,
    error_msg: Optional[str] = None,
) -> Response:
    """Get information of a specific Dynamic Recipient List for a given project.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): ID of the Dynamic Recipient List
        project_id (str): ID of the project
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.get(
        url=f'{connection.base_url}/api/dynamicRecipientLists/{id}',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg='Error creating new subscription.')
def create_subscription(connection, project_id, body, fields=None, error_msg=None):
    """Create a new subscription.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        project_id (str): ID of the project
        body: JSON-formatted body of the new subscription, example;
        {"name": "My new subscription",
        "allowDeliveryChanges": false,
        "allowPersonalizationChanges": false,
        "allowUnsubscribe": false,
        "sendNow": false,
        "schedules": [
            {
            "id": "string"
            }
        ],
        "contents": [
            {
            "id": "string",
            "type": "report",
            "personalization": {
                "compressed": false,
                "formatMode": "DEFAULT",
                "viewMode": "DEFAULT",
                "formatType": "PLAIN_TEXT",
                "delimiter": "string",
                "bursting": {
                "slicingAttributes": [
                    "string"
                ],
                "addressAttributeId": "string",
                "deviceId": "string",
                "formId": "string"
                },
                "prompt": {
                "enabled": false,
                "instanceId": "string"
                }
            }
            }
        ],
        "recipients": [
            {
            "id": "string",
            "includeType": "TO"
            }
        ],
        "delivery": {
            "mode": "EMAIL",
            "expiration": "2020-09-21T10:35:01.911Z",
            "contactSecurity": true,
            "email": {
            "subject": "string",
            "message": "string",
            "filename": "string",
            "spaceDelimiter": "string",
            "includeLink": true,
            "includeData": true,
            "sendToInbox": true,
            "overwriteOlderVersion": true,
            "zip": {
                "filename": "string",
                "password": "string",
                "passwordProtect": true
            }
            },
            "file": {
            "filename": "string",
            "spaceDelimiter": "string",
            "burstSubFolder": "string",
            "zip": {
                "filename": "string",
                "password": "string",
                "passwordProtect": true
            }
            },
            "printer": {
            "copies": 0,
            "rangeStart": 0,
            "rangeEnd": 0,
            "collated": true,
            "orientation": "PORTRAIT",
            "usePrintRange": true
            },
            "ftp": {
            "spaceDelimiter": "string",
            "filename": "string",
            "zip": {
                "filename": "string",
                "password": "string",
                "passwordProtect": true
            }
            },
            "cache": {
            "cacheType": "RESERVED",
            "shortcutCacheFormat": "RESERVED"
            },
            "mobile": {
            "clientType": "RESERVED",
            "deviceId": "string",
            "doNotCreateUpdateCaches": true,
            "overwriteOlderVersion": true,
            "reRunHl": true
            },
            "historyList": {
            "deviceId": "string",
            "doNotCreateUpdateCaches": true,
            "overwriteOlderVersion": true,
            "reRunHl": true
            },
            "iserverExpiration": "string"
        }
        }
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.post(
        url=f'{connection.base_url}/api/subscriptions',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg='Error creating new Dynamic Recipient List.')
def create_dynamic_recipient_list(
    connection: "Connection",
    project_id: str,
    body: dict,
    fields: Optional[list[str]] = None,
    error_msg: Optional[str] = None,
) -> Response:
    """Create a new subscription.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        project_id (str): ID of the project
        body: JSON-formatted body of the new subscription
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.post(
        url=f'{connection.base_url}/api/dynamicRecipientLists',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


def remove_subscription(
    connection, subscription_id, project_id, error_msg=None, exception_type=None
):
    """Remove (Unsubscribe) the subscription using subscription id.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        subscription_id (str): ID of the subscription
        project_id (str): ID of the project
        error_msg (str, optional): Customized error message.
        exception_type (Exception): Instance of Exception or Warning class

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    response = connection.delete(
        url=connection.base_url + '/api/subscriptions/' + subscription_id,
        headers={'X-MSTR-ProjectID': project_id},
    )
    if not response.ok:
        if error_msg is None:
            error_msg = f"Error unsubscribing Subscription {subscription_id}"
        if exception_type is None:
            response_handler(response, error_msg)
        else:
            exception_handler(error_msg, exception_type)
    return response


def remove_dynamic_recipient_list(
    connection: "Connection",
    id: str,
    project_id: str,
    error_msg: Optional[str] = None,
    exception_type: Optional[Exception] = None,
) -> Response:
    """Delete a Dynamic Recipient List.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): ID of the Dynamic Recipient List
        project_id (str): ID of the project
        error_msg (str, optional): Customized error message.
        exception_type (Exception): Instance of Exception or Warning class

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    response = connection.delete(
        url=connection.base_url + '/api/dynamicRecipientLists/' + id,
        headers={'X-MSTR-ProjectID': project_id},
    )
    if not response.ok:
        if error_msg is None:
            error_msg = f"Error deleting Dynamic Recipient List id: {id}"
        if exception_type is None:
            response_handler(response, error_msg)
        else:
            exception_handler(error_msg, exception_type)
    return response


@ErrorHandler(err_msg='Error updating subscription {subscription_id}')
def update_subscription(
    connection, subscription_id, project_id, body, fields=None, error_msg=None
):
    """Updates a subscription.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        subscription_id (str): ID of the subscription
        project_id (str): ID of the project
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.put(
        url=f'{connection.base_url}/api/subscriptions/{subscription_id}',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg='Error updating Dynamic Recipient List {list_id}')
def update_dynamic_recipient_list(
    connection: "Connection",
    id: str,
    project_id: str,
    body: dict,
    fields: Optional[list[str]] = None,
    error_msg: Optional[str] = None,
) -> Response:
    """Updates a Dynamic Recipient List.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): ID of the Dynamic Recipient List
        project_id (str): ID of the project
        body: JSON-formatted body of the new subscription
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.put(
        url=f'{connection.base_url}/api/dynamicRecipientLists/{id}',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg='Error getting recipients list.')
def available_recipients(
    connection,
    project_id,
    body,
    delivery_type,
    offset=0,
    limit=-1,
    fields=None,
    error_msg=None,
):
    """Get a list of available recipients in shared list, for a given content
    and delivery type, within a given project.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (str): ID of the project
        body (json): {
                    "contents": [
                        {
                        "id": "string",
                        "type": "REPORT"
                        }
                    ]
                    }
        delivery_type (str): Type of the delivery [EMAIL / FILE / PRINTER /
            HISTORY_LIST / CACHE / MOBILE / FTP]
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.


    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.post(
        url=f'{connection.base_url}/api/subscriptions/recipients/results',
        params={
            'fields': fields,
            'deliveryType': delivery_type,
            'offset': offset,
            'limit': limit,
        },
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg='Error getting available bursting attributes list.')
def bursting_attributes(
    connection, project_id, content_id, content_type, fields=None, error_msg=None
):
    """Get a list of available attributes for bursting feature, for a given
    content, within a given project. This endpoint returns the name, ID, and
    other information about available attributes.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (str): ID of the project
        content_id (str): ID of the content
        content_type (str): Type of the content [REPORT, DOCUMENT]
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.get(
        url=f'{connection.base_url}/api/subscriptions/bursting',
        params={'fields': fields, 'contentId': content_id, 'contentType': content_type},
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg='Error sending subscription {subscription_id}')
def send_subscription(
    connection, subscription_id, project_id, body, fields=None, error_msg=None
):
    """Send the existing subscription immediately.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        subscription_id (str): ID of subscription
        project_id (str): ID of the project
        body (dict): body of the request
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Customized error message.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.get(
        url=f'{connection.base_url}/api/subscriptions/{subscription_id}/send',
        params={'fields': fields},
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )



"""This is the demo script to show how administrator can manage reports.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.project_objects.content_cache import ContentCache
from mstrio.project_objects.report import Report, list_reports

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Cache management
# List all report caches
cache_list = Report.list_caches(conn)

# list all reports
reports = list_reports(connection=conn)

# Define a variable which can be later used in a script
REPORT_ID = $report_id  # Insert ID of report here

# When extracting contents of a report instance, new cache is created
# (for projects where caching for the report is enabled)
Report(id=REPORT_ID, connection=conn).to_dataframe()

# List all report caches again (there is a new cache)
updated_list = Report.list_caches(conn)

# Define variables which can be later used in a script
REPORT_CACHE_ID_1 = $report_cache_id_1  # Insert ID of report cache on which you want to perform actions
REPORT_CACHE_ID_2 = $report_cache_id_2  # Insert ID of report cache on which you want to perform actions

# List report caches with different conditions
list_of_all_caches = Report.list_caches(conn)

# Define variables which can be later used in the script
PROJECT_ID = $project_id  # Insert project ID here
NODE_NAME = $node_name  # Insert name of node or list of names on which you want to perform actions

# List report caches in a project
list_of_caches_for_project = Report.list_caches(connection=conn, project_id=PROJECT_ID)

# List report caches in a node
list_of_caches_for_node = Report.list_caches(connection=conn, nodes=NODE_NAME)

# List report caches by id
list_of_one_cache = Report.list_caches(connection=conn, id=REPORT_CACHE_ID_1)

# List report caches as dictionary
list_of_caches_as_dicts = Report.list_caches(connection=conn, to_dictionary=True)

# Define variables which can be later used in a script
DB_CONNECTION_ID = $db_connection_id  # Insert ID of DB connection you want to include in your functions
DB_LOGIN_ID = $db_login_id  # Insert ID of DB login you want to include in your functions

# List report caches for database connection
list_of_caches_for_db_connection = Report.list_caches(
    connection=conn, db_connection_id=DB_CONNECTION_ID
)

# List report caches for database login
list_of_caches_for_db_login = Report.list_caches(
    connection=conn, db_login_id=DB_LOGIN_ID
)

# Define variables which can be later used in a script
USER_NAME = $user_name  # Insert name of the owner of some cache
WH_TABLE_ID = $wh_table_id  # Insert warehouse table ID that some cache is using

# List report caches by owner
list_of_caches_for_owner = Report.list_caches(connection=conn, owner=USER_NAME)

# List report caches for warehouse table
list_of_caches_for_wh_table = Report.list_caches(connection=conn, wh_tables=WH_TABLE_ID)

# List report caches by status
list_of_caches_loaded = Report.list_caches(connection=conn, status='loaded')

# Get single report cache by its id
report_cache = ContentCache(connection=conn, id=REPORT_CACHE_ID_1)

# Listing properties of content cache
properties = report_cache.list_properties()

# Unload content cache
report_cache.unload()

# Check report cache status
status = report_cache.status
print(status)

# Load content cache
report_cache.load()

# Delete content cache with confirmation prompt
report_cache.delete()

# Delete content cache without confirmation prompt
report_cache.delete(force=True)

# Refresh content cache
report_cache.fetch()

# Managing multiple specific caches
# Load multiple caches
ContentCache.load_caches(conn, cache_ids=[REPORT_CACHE_ID_1, REPORT_CACHE_ID_2])

# Unload multiple caches
ContentCache.unload_caches(conn, cache_ids=[REPORT_CACHE_ID_1, REPORT_CACHE_ID_2])

# Delete multiple caches with confirmation prompt
ContentCache.delete_caches(conn, cache_ids=[REPORT_CACHE_ID_1, REPORT_CACHE_ID_2])

# Delete multiple caches without confirmation prompt
ContentCache.delete_caches(conn, cache_ids=[REPORT_CACHE_ID_1, REPORT_CACHE_ID_2], force=True)

# Managing all caches meeting the criteria
# Load all report caches
Report.load_all_caches(conn)

# Load all report caches owned by specific user
Report.load_all_caches(connection=conn, owner=USER_NAME)

# Unload all report caches
Report.unload_all_caches(conn)

# Unload all report caches owned by specific user
Report.unload_all_caches(connection=conn, owner=USER_NAME)

# Delete all report caches
Report.delete_all_caches(conn)  # with confirmation prompt
Report.delete_all_caches(conn, force=True)  # without confirmation prompt

# Delete all report caches owned by specific user
Report.delete_all_caches(connection=conn, owner=USER_NAME)  # with confirmation prompt
Report.delete_all_caches(connection=conn, owner=USER_NAME, force=True)  # without confirmation prompt

# Delete all loaded report caches
Report.delete_all_caches(connection=conn, status='loaded')  # with confirmation prompt
Report.delete_all_caches(connection=conn, status='loaded', force=True)  # without confirmation prompt



import json


def alter_conn_resp(response):
    response_json = response.json()
    response_json = alter_conn_json(response_json)
    response.encoding, response._content = 'utf-8', json.dumps(response_json).encode(
        'utf-8'
    )
    return response


def alter_conn_list_resp(response):
    response_json = response.json()
    for conn in response_json["connections"]:
        alter_conn_json(conn)
    response.encoding, response._content = 'utf-8', json.dumps(response_json).encode(
        'utf-8'
    )
    return response


def alter_conn_json(response_json):
    response_json['datasource_login'] = response_json["database"]["login"]
    response_json['database_type'] = response_json["database"]["type"]
    response_json['database_version'] = response_json["database"]["version"]
    del response_json["database"]
    return response_json


def alter_instance_resp(response):
    response_json = response.json()
    response_json = alter_instance_json(response_json)
    response.encoding, response._content = 'utf-8', json.dumps(response_json).encode(
        'utf-8'
    )
    return response


def alter_instance_list_resp(response):
    response_json = response.json()
    for ds in response_json["datasources"]:
        alter_instance_json(ds)
    response.encoding, response._content = 'utf-8', json.dumps(response_json).encode(
        'utf-8'
    )
    return response


def alter_instance_json(response_json):
    response_json['datasource_connection'] = response_json["database"].get("connection")
    response_json['database_type'] = response_json["database"].get("type")
    response_json['database_version'] = response_json["database"].get("version")
    response_json['primary_datasource'] = response_json["database"].get(
        "primaryDatasource"
    )
    response_json['data_mart_datasource'] = response_json["database"].get(
        "dataMartDatasource"
    )
    del response_json["database"]
    return response_json


def alter_patch_req_body(op_dict, initial_path, altered_path):
    if op_dict["path"] == initial_path:
        op_dict["path"] = altered_path
        op_dict["value"] = (
            op_dict["value"]
            if isinstance(op_dict["value"], str)
            else op_dict["value"].get("id")
        )
    return op_dict


def get_objects_id(obj, obj_class):
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, obj_class):
        return obj.id
    return None



import csv
from enum import Enum
import inspect
import logging
from os.path import join as joinpath
from pprint import pprint
from sys import version_info
from typing import (
    Any,
    Callable,
    Optional,
    TYPE_CHECKING,
    TypeVar,
)

from pandas import DataFrame
import humps

from mstrio import config
from mstrio.api import objects
from mstrio.api.exceptions import VersionException
from mstrio.connection import Connection
from mstrio.types import ExtendedType, ObjectSubTypes, ObjectTypes
from mstrio.utils import helper
from mstrio.utils.acl import ACE, ACLMixin, Rights
from mstrio.utils.dependence_mixin import DependenceMixin
from mstrio.utils.exceptions import NotSupportedError
from mstrio.utils.helper import rename_dict_keys
from mstrio.utils.time_helper import (
    bulk_str_to_datetime,
    DatetimeFormats,
    map_str_to_datetime,
)

if TYPE_CHECKING:
    from mstrio.object_management import Folder
    from mstrio.server import Project

logger = logging.getLogger(__name__)

T = TypeVar("T")


class EntityBase(helper.Dictable):
    """This class is for objects that do not have a specified MSTR type.

    Class attributes:
    _OBJECT_TYPE (ObjectTypes): MSTR Object type defined in ObjectTypes
    _OBJECT_SUBTYPES (list[ObjectSubTypes] | None): MSTR Object subtype defined
        in ObjectSubTypes. It contains a list of subtypes, which the class
        supports. Used in fetch(), to verify whether the retrieved object
        is supported. None means that subtype won't be verified.
    _REST_ATTR_MAP (Dict[str, str]): A dictionary whose keys are names of
        fields of an HTTP response from the server, and values are their
        respective Python API mappings. This dictionary is mainly used when
        overriding `_init_variables` function to convert field names to
        attribute names before the object's attributes initialization. For
        example, if a value of the `job_type` field should be stored in a 'type'
        attribute, one can specify this dictionary as: {"job_type" : "type"} and
        override `_init_variables` method to implement this mapping such as:
         ```
            kwargs = self._rest_to_python(kwargs)
            self._AVAILABLE_ATTRIBUTES.update({key: type(val) for key, val
                in kwargs.items()})
            # init logic follows
        ```
    _API_GETTERS (dict[str | tuple, Callable]): A dictionary whose keys
        are either a name of an attribute (as a string) or names of attributes
        (as a tuple), and values are the REST API wrapper functions used to
        fetch related data from a server. For example:
        {'id': objects.get_object_info} or
        {('id', 'name', 'description): objects.get_object_info}
    _FROM_DICT_MAP (dict[str, Callable]): A dictionary whose keys are
        attribute's name and values are the attribute's type. This mapping is
        required to determine a proper composite form in which a value will be
        stored (such as an Enum, list of Enums etc.). Therefore, only attributes
        with composite and not primitive data types should be included here.
    _AVAILABLE_ATTRIBUTES (dict[str, type]): A dictionary which keys are
        object's attribute names as strings, and which values are types of
        attributes value. It is used for validating attribute's type during the
        process of properties update. This dictionary is created from keyword
        arguments passed to an object's constructor.
    _API_PATCH (dict[tuple[str], tuple[Callable | str]]): A dictionary
        whose keys are tuples with an object's attribute names as strings, and
        values are tuples with two elements: first element is a REST API wrapper
        function used to update the object's properties, and the second element
        is a definition (as a string) how this update should be performed. For
        example:
        {('name', 'description', 'abbreviation'):
            (objects.update_object, 'partial_put')}
    _PATCH_PATH_TYPES (dict[str, type]): A dictionary whose keys are names of
        object's attributes and values are attribute types. Used to
        validate correctness of a new value during the process of properties
        update. For example: {'id': str, 'enabled': bool}
    _API_DELETE(Callable): A function which is used to delete an object from the
        server. Defaults to staticmethod(objects.delete_object).


    Instance attributes:
        connection (Connection): A MicroStrategy connection object
        id (str): Object's ID
        type (ObjectTypes): MicroStrategy Object Type
        name (str): Object's name
        _altered_properties (dict): This is a private attribute which is used to
            track and validate local changes to the object. It is used whenever
            an attribute is trying to be set on the object (see `__setattr__`).
            It is also used to determine which properties should be updated on
            a server (see `update_properties`)
        _fetched_attributes (set): This is a private attribute which is used to
            track which attributes have been already fetched from the server.
            The logic of `__getattribute__` method is based on information
            in this set.
        subtype (ObjectSubTypes): The enumeration constant used to specify the
            object's sub-type, which reveals more specific information than the
            object's type.
        ext_type (ExtendedType): A part of an object stored in the metadata.
            It is used is to store additional custom information that can be
            used to achieve new functionality. For example, to display a report
            as a map and save it as part of the Report object, you can create an
            extended property for displaying a report as a map and store it on
            a server in the object's properties set.
        date_created (DatetimeFormats): The object's creation time.
        date_modified (DatetimeFormats): The object's last modification time.
        version (str): The object's version ID. Used to compare if two MSTR
            objects are identical. If both objects IDs and objects version IDs
            are the same, MSTR Object Manager determines that objects as
            'Exists Indentically'. Otherwise, if their IDs match but their
            version IDs mismatch, MSTR Object Manager determines that objects
            'Exists Differently'.
        owner (User): The object's owner information.
        icon_path (str): A path to a location where the object's icon is stored.
        view_media (int): The enumeration constant used to represent the default
            mode of a RSD or a dossier, and available modes of a RSD or a
            dossier.
        ancestors (list[dict]): A list of the object's ancestor folders.
        certified_info (CertifiedInfo): The object's certification status,
            time of certification, and information about the certifier
            (currently only for document and report)
        acg (Rights): The enumeration constant used to specify the access
            granted attribute of the object.
        acl (list[ACE]): A list of permissions on the object so that users, or
            user groups have control over individual objects in the system.
            Those permissions decide whether or not a user can perform a
            particular class of operations on a particular object. For example,
            a user may have permissions to view and execute a report , but
            cannot modify the report definition or delete the report.
        hidden (bool): Determines whether the object is hidden on a server.
        project_id (str): The ID of a project in which the object resides.
        comments (list[str]): Custom user's comments related to the object
        target_info (dict): ?
    """

    _OBJECT_TYPE: ObjectTypes = (
        ObjectTypes.NONE
    )  # MSTR object type defined in ObjectTypes
    _OBJECT_SUBTYPES: list[
        ObjectSubTypes
    ] | None = None  # None means subtype won't be verified.
    _REST_ATTR_MAP: dict[str, str] = {}
    _API_GETTERS: dict[str | tuple, Callable] = {}
    _FROM_DICT_MAP: dict[str, Callable] = {
        'type': ObjectTypes
    }  # map attributes to Enums and Composites
    _AVAILABLE_ATTRIBUTES: dict[str, type] = {}  # fetched on runtime from all Getters
    _API_PATCH: dict[tuple[str], tuple[Callable | str]] = {}
    _PATCH_PATH_TYPES: dict[str, type] = {}  # used in update_properties method
    _API_DELETE: Callable = staticmethod(objects.delete_object)

    def __init__(self, connection: Connection, object_id: str, **kwargs) -> None:
        self._init_variables(connection=connection, id=object_id, **kwargs)
        if config.fetch_on_init and self._find_func("id") is not None:
            self.fetch("id")
        if config.verbose:
            logger.info(self)

    def _init_variables(self, **kwargs) -> None:
        """Initializes object's attributes from kwargs. This method is often
        overridden by subclasses of Entity / EntityBase to abstract the process
        of initialization as it automatically sets attributes mutual for all
        entities, adds information about object-specific attributes to
        `_AVAILABLE_ATTRIBUTES` dictionary and creates a `_altered_properties`
        dictionary which is used to track and validate any local
        changes to the object. `_altered_properties` is later used to
        properly update object's properties on a server.

        Note: attributes not accepted by any implementation of this function
            in the inheritance chain will be disregarded.
        """
        kwargs = self._rest_to_python(kwargs)
        # create _AVAILABLE_ATTRIBUTES map
        self._AVAILABLE_ATTRIBUTES.update(
            {key: type(val) for key, val in kwargs.items()}
        )
        self._connection = kwargs.get("connection")
        self._id = kwargs.get("id")
        self._type = (
            self._OBJECT_TYPE
            if 'type' in self._FROM_DICT_MAP
            else kwargs.get("type", self._OBJECT_TYPE)
        )
        self.name = kwargs.get("name")
        self._altered_properties = dict()

    def fetch(self, attr: Optional[str] = None) -> None:  # NOSONAR
        """Fetch the latest object's state from the I-Server.

        Note:
            This method can overwrite local changes made to the object.

        Args:
            attr (Optional[str]): Attribute name to be fetched. If not specified
            it will use all getters specified in `_API_GETTERS` dictionary.
            Defaults to None.
        Raises:
            ValueError: If `attr` cannot be fetched.
        """
        functions = self._API_GETTERS  # by default fetch all endpoints

        if (
            attr is not None
        ):  # if attr is specified fetch endpoint matched to the attribute name
            function = self._find_func(attr)
            if not function:
                raise ValueError(
                    f"The attribute `{attr}` cannot be fetched for this object"
                )
            else:
                functions = {
                    attr: func for attr, func in functions.items() if func == function
                }

        for key, func in functions.items():  # call respective API getters
            param_value_dict = auto_match_args_entity(func, self)

            response = func(**param_value_dict)

            if response:
                json = (
                    response if isinstance(response, (dict, list)) else response.json()
                )

                if self._OBJECT_SUBTYPES and (subtype := json['subtype']):
                    self._check_object_subtype(subtype)
                if type(json) == dict:
                    object_dict = {
                        key if isinstance(key, str) and len(json) == 1 else k: v
                        for k, v in json.items()
                    }
                    self._set_object_attributes(**object_dict)
                elif type(json) == list:
                    self._set_object_attributes(**{key: json})

            # keep track of fetched attributes
            self._add_to_fetched(key)

    @classmethod
    def _check_object_subtype(cls, subtype: int) -> None:
        """Check if subtype is supported by a class.

        Args:
            subtype(int): number representing object's subtype.

        Returns:
            None

        Raises:
            NotSupportedError: If suptype is not in _OBJECT_SUBTYPES list.
        """
        if cls._OBJECT_SUBTYPES and not cls._is_subtype_supported(subtype):
            raise NotSupportedError(
                f"Objects with subtype {ObjectSubTypes(subtype) !r} are not supported."
            )

    @classmethod
    def _is_subtype_supported(cls, subtype: int) -> bool:
        """Check if subtype is supported by a class.

        Args:
            subtype (int): number representing object's subtype.

        Returns:
            True if subtype is supported by class.
            False if subtype is not supported.
        """
        return subtype in [item.value for item in cls._OBJECT_SUBTYPES]

    def _add_to_fetched(self, keys: str | tuple) -> None:
        """Adds name/-s of attribute/-s to the `_fetched_attributes` set.

        Args:
            keys (str | tuple): Name, or tuple with names of attributes.
        """
        if isinstance(keys, str):
            keys = [keys]
        for key in keys:
            key = key[1:] if key.startswith("_") else key
            self._fetched_attributes.add(key)

    @classmethod
    def _find_func(cls, attr: str) -> Optional[Callable]:
        """Searches cls._API_GETTERS dictionary for a REST API wrapper function
        used to fetch `attr` from a server.

        Args:
            attr (str): Name of an attribute to be fetched from the server.

        Raises:
            TypeError: `attr` not passed as a string.
            ValueError: cls._API_GETTERS dictionary was not specified
            properly.

        Returns:
            Optional[Callable]: REST API wrapper function corresponding with
                `attr` parameter, else None.
        """
        if not isinstance(attr, str):
            raise TypeError("`attr` parameter has to be of type str")

        for attributes, func in cls._API_GETTERS.items():
            if isinstance(attributes, str):
                if attr == attributes:
                    return func
            elif isinstance(attributes, tuple):
                if attr in attributes:
                    return func
            else:
                raise ValueError
        return None

    @classmethod
    def _rest_to_python(cls, response: dict) -> dict:
        """Map REST API field names to Python API field names as specified in
        cls._REST_ATTR_MAP.

        Args:
            response (dict): A dictionary representing an HTTP response.

        Returns:
            dict: A dictionary with field names converted to Python API names.
        """

        return rename_dict_keys(response, cls._REST_ATTR_MAP)

    @classmethod
    def _python_to_rest(cls, request_body: dict) -> dict:
        """Map Python API field names to REST API field names as specified in
        cls._REST_ATTR_MAP.

        Args:
            request_body (dict): A dictionary representing an HTTP request body.

        Returns:
            dict: A dictionary with field names converted to REST API names.
        """
        body = request_body.copy()
        for rest_name, python_name in cls._REST_ATTR_MAP.items():
            if python_name in body:
                body[rest_name] = body.pop(python_name)
        return body

    def __compose_val(self, key: str, val: Any) -> Any:
        """Converts a value to a correct composite form such as Datetime, Enum,
        list of Enums etc. Information about the correct form of the value is
        stored in the `self._FROM_DICT_MAP`

        Args:
            key (str): An object's attribute name used to determine the correct
                form of a value.
            val (Any): A value which will be converted to a correct form
        Returns:
            Any: A value converted to a correct form.
        """
        if key in self._FROM_DICT_MAP:
            # determine which constructor will be used
            if isinstance(self._FROM_DICT_MAP[key], DatetimeFormats):
                val = map_str_to_datetime(key, val, self._FROM_DICT_MAP)
            elif isinstance(self._FROM_DICT_MAP[key], type(Enum)):
                val = self._FROM_DICT_MAP[key](val)
            elif isinstance(self._FROM_DICT_MAP[key], list) and val is not None:
                if isinstance(self._FROM_DICT_MAP[key][0], type(Enum)):
                    val = [self._FROM_DICT_MAP[key][0](v) for v in val]
                else:
                    val = [
                        self._FROM_DICT_MAP[key][0](
                            source=v, connection=self.connection
                        )
                        for v in val
                    ]
            elif val == {}:
                return None
            else:
                val = self._FROM_DICT_MAP[key](source=val, connection=self.connection)

        return val

    def _set_object_attributes(self, **kwargs) -> None:
        """Set object's attributes programmatically by providing keyword args.
        Support ENUMs and creating composites. This function inspects whether
        the attribute was set as a property, and if yes it stores the attribute
        with '_' prefix"""

        object_info = helper.camel_to_snake(kwargs, whitelist=self._KEEP_CAMEL_CASE)
        object_info = self._rest_to_python(object_info)

        # determine which attributes should be private
        properties = {
            elem[0]
            for elem in inspect.getmembers(
                self.__class__, lambda x: isinstance(x, property)
            )
        }

        for key, val in object_info.items():  # type: ignore
            # update _AVAILABLE_ATTRIBUTES map
            self._AVAILABLE_ATTRIBUTES.update({key: type(val)})

            # if self is a composite, create component instance
            val = self.__compose_val(key, val)

            # check if attr is read-only and if yes return '_' version of it
            if key in properties:
                key = "_" + key

            setattr(self, key, val)

    def list_properties(self) -> dict:
        """Fetches all attributes from the server and converts all properties of
        the object to a dictionary.

        Returns:
            dict: A dictionary which keys are object's attribute names, and
                which values are object's attribute values.
        """
        if hasattr(self, "_API_GETTERS"):  # fetch attributes not loaded on init
            attr = [attr for attr in self._API_GETTERS.keys() if isinstance(attr, str)]
            for a in attr:
                try:
                    getattr(self, a)
                except VersionException:
                    pass

        properties = inspect.getmembers(
            self.__class__, lambda x: isinstance(x, property)
        )
        properties = {elem[0]: elem[1].fget(self) for elem in properties}
        attributes = {
            key: val for key, val in vars(self).items() if not key.startswith('_')
        }
        attributes = {**properties, **attributes}

        return {
            key: attributes[key]
            for key in sorted(attributes, key=helper.sort_object_properties)
        }

    def to_dataframe(self) -> DataFrame:
        """Converts all properties of the object to a dataframe.

        Returns:
            DataFrame: A `DataFrame` object containing object properties.
        """
        return DataFrame.from_dict(
            self.list_properties(), orient='index', columns=['value']
        )

    def print(self) -> None:
        """Pretty Print all properties of the object."""
        if version_info.major >= 3 and version_info.minor >= 8:
            pprint(self.list_properties(), sort_dicts=False)  # type: ignore
        else:
            pprint(self.list_properties())

    @classmethod
    def from_dict(
        cls: T,
        source: dict[str, Any],
        connection: Connection,
        to_snake_case: bool = True,
    ) -> T:
        """Overrides `Dictable.from_dict()` to instantiate an object from
            a dictionary without calling any additional getters.

        Args:
            cls (T): Class (type) of an object that should be created.
            source (dict[str, Any]): a dictionary from which an object will be
                constructed.
            connection (Connection): A MicroStrategy Connection object.
            to_snake_case (bool, optional): Set to True if attribute names
            should be converted from camel case to snake case. Defaults to True.

        Returns:
            T: An object of type T.
        """

        obj = cls.__new__(cls)  # Does not call __init__
        object_source = helper.camel_to_snake(source) if to_snake_case else source
        obj._init_variables(connection=connection, **object_source)
        return obj

    def __str__(self) -> str:
        """Overrides default `__str__` method to provide more informative
        description of an object.

        Returns:
            str: A formatted string consisting of object's type, name and id.
        """
        if self.__dict__.get("name"):
            return "{} object named: '{}' with ID: '{}'".format(
                self.__class__.__name__, self.name, self.id
            )
        else:
            return f"{self.__class__.__name__} object with ID: '{self.id}'"

    def __repr__(self):
        param_value_dict = auto_match_args_entity(
            self.__init__, self, exclude=['self'], include_defaults=False
        )
        params_list = []
        for param, value in param_value_dict.items():
            if param == "connection" and isinstance(value, Connection):
                params_list.append("connection")
            else:
                params_list.append(f"{param}={repr(value)}")
        formatted_params = ", ".join(params_list)
        return f"{self.__class__.__name__}({formatted_params})"

    def __hash__(self):
        return hash((self.id, self._OBJECT_TYPE.value))

    def __eq__(self, other):
        """Equals operator to compare if an entity is equal to another object.

        We define 2 Entities as being equal if:
            1. the other object is an Entity and also shares the same id
            2. the other object is a string and is equal to either the name or
               id of the Entity
        This allows us to search for Entities in a collection by specifying its
        id, name, or another Entity itself. For example:
            if "user" in users: ...
            if "28ECA8BB11D5188EC000E9ABCA1B1A4F" in users: ...
        """
        if isinstance(other, str):
            return self.id == other or self.name == other
        elif isinstance(other, type(self)):
            return self.id == other.id
        elif isinstance(other, dict):
            # check all items in other dict equal entity object attributes
            return all(
                (
                    getattr(self, k) == v if hasattr(self, k) else False
                    for k, v in other.items()
                )
            )
        else:
            return NotImplemented  # don't attempt to compare against unrelated types

    @classmethod
    def to_csv(  # NOSONAR
        cls: T,
        objects: T | list[T],
        name: str,
        path: Optional[str] = None,
        properties: Optional[list[str]] = None,
    ) -> None:
        """Exports MSTR objects to a csv file.

        Optionally, saves only the object properties specified in the properties
        parameter.

        Args:
            objects (T | list[T]): List of objects of the same type that
            will be exported.
            name (str): The name of the csv file ending with '.csv'
            path (Optional[str], optional): A path to the directory where the
                file will be saved. Defaults to None.
            properties (Optional[list[str]], optional): A list of object's
                attribute names that should be included in the exported file.
                Defaults to None.

        Raises:
            TypeError: If `objects` is not of type `T` or list of type `T`
            objects.
        """
        file = joinpath(path, name) if path else name
        list_of_objects = []
        if not name.endswith('.csv'):
            msg = (
                "The file extension is different than '.csv', please note that using a "
                "different extension might disrupt opening the file correctly."
            )
            helper.exception_handler(msg, exception_type=Warning)

        if not isinstance(objects, list):
            objects = [objects]
        for obj in objects:
            if not isinstance(obj, cls):
                helper.exception_handler(
                    (
                        f"Object '{obj}' of type '{type(obj)}' is not supported."
                        f"Objects should be of type {cls.__name__} or "
                        f"list of {cls.__name__}."
                    ),
                    exception_type=TypeError,
                )
            if properties:
                list_of_objects.append({key: getattr(obj, key) for key in properties})
            else:
                list_of_objects.append(obj.list_properties())

        with open(file, 'w') as f:
            fieldnames = list_of_objects[0].keys()
            w = csv.DictWriter(f, fieldnames=fieldnames)
            w.writeheader()
            w.writerows(list_of_objects)
        if config.verbose:
            logger.info(f"Object exported successfully to '{file}'")

    def update_properties(self) -> None:
        """Save compatible local changes of the object attributes to the
        I-Server. Changes are retrieved from the `self._altered_properties`
        dictionary. After the process of update has finished,
        `self._altered_properties` is cleared. For this method to work
        properly, you must override the `_alter_properties()` method in a
        subclass.

        Raises:
            requests.HTTPError: If I-Server raises exception
        """
        changes = {k: v[1] for k, v in self._altered_properties.items()}
        self._alter_properties(**changes)
        self._altered_properties.clear()

    def __partial_put_body(self, attrs: tuple, camel_properties: dict) -> dict:
        """Prepares the 'partial put' update request's body.

        Args:
            attrs (tuple): A tuple consisting of name/-s of attributes.
                Corresponds to keys in a `_API_PATCH` dictionary.
            camel_properties (dict): A dictionary representation of an object's
                attribute names (in camel case) and values.

        Returns:
            dict: A dictionary used as a PUT request body.
        """
        body = {}
        for name, value in camel_properties.items():
            snake_case_name = humps.decamelize(name)
            if snake_case_name in attrs:
                value = self._unpack_objects(name, value, camel_case=True)
                body[name] = self._validate_type(snake_case_name, value)
        return body

    def __put_body(self, attrs: tuple, properties: dict) -> dict:
        """Prepares the 'put' update request's body.

        Args:
            attrs (tuple): A tuple consisting of name/-s of attributes.
                Corresponds to keys in a `_API_PATCH` dictionary.
            properties (dict): A dictionary representation of an object's
                attribute names (in snake case) and values.

        Returns:
            dict: A dictionary used as a PUT request body.
        """
        for name, value in properties.items():
            if name in attrs:
                value = self._unpack_objects(name, value, camel_case=True)
                setattr(self, name, self._validate_type(name, value))
        return self.to_dict()

    def __patch_body(self, attrs: tuple, camel_properties: dict, op: str) -> dict:
        """Prepares the 'patch' update request's body.

        Args:
            attrs (tuple): A tuple consisting of name/-s of attributes.
                Corresponds to keys in a `_API_PATCH` dictionary.
            camel_properties (dict): A dictionary representation of an object's
                attribute names (in snake case) and values.
            op (str): An operation type. Possible values are add, replace,
                remove, removeElement, addElement, removeElements or
                addElements.

        Returns:
            dict: A dictionary used as a PATCH request body.
        """

        body = {"operationList": []}
        for name, value in camel_properties.items():
            snake_case_name = humps.decamelize(name)
            if snake_case_name in attrs:
                value = self._unpack_objects(name, value, camel_case=True)
                body['operationList'].append(
                    {
                        "op": op,
                        "path": f"/{name}",
                        "value": self._validate_type(snake_case_name, value),
                    }
                )
        return body

    def _send_proper_patch_request(
        self, properties: dict, op: str = 'replace'
    ) -> list[bool]:
        """Internal method to update objects with the specified patch wrapper.
        Used for adding and removing objects from nested properties of an
        object like memberships. Automatically converts `properties` keys from
        a snake case to a camel case. The update's type is specified in the
        `self._API_PATCH` dictionary and be described as follows:
        - 'patch' : Used when the REST API provides a PATCH endpoint to update
            the object.
        - 'put': Used when the REST API requires that a body of a request
            consists of all the object's attributes.
        - 'partial_put': Used when the REST API requires that a body of a
            request consists only of attributes that are being updated. Although
            it resembles and imitates PATCH request, a PUT HTTP method is used,
            hence its name.

        Args:
            properties (dict): A dictionary of required changes
            op (str): An operation type. Possible values are add, replace,
                remove, removeElement, addElement, removeElements or
                addElements. Defaults to 'replace'.

        Returns:
            A list of successful or unsuccessful requests.

        Raises:
            NotSupportedError: If there is no information in `self._API_PATCH`
                dictionary about the update's type ('partial_put', 'put' or
                'patch').
        """
        changed = []
        camel_properties = helper.snake_to_camel(
            properties, whitelist=self._KEEP_CAMEL_CASE
        )
        for attrs, (func, func_type) in self._API_PATCH.items():
            if func_type == 'partial_put':
                translated_properties = self._python_to_rest(camel_properties)
                body = self.__partial_put_body(attrs, translated_properties)
            elif func_type == 'put':  # Update using the generic update_object()
                body = self.__put_body(attrs, properties)
                body = self._python_to_rest(body)
            elif func_type == 'patch':
                translated_properties = self._python_to_rest(camel_properties)
                body = self.__patch_body(attrs, translated_properties, op)
            else:
                msg = (
                    f"{func} function is not supported by `_send_proper_patch_request`"
                )
                raise NotSupportedError(msg)

            if not body:
                continue
            # send patch request from the specified update wrapper
            param_value_dict = auto_match_args_entity(func, self, exclude=["body"])
            param_value_dict['body'] = body
            response = func(**param_value_dict)

            if response:
                changed.append(True)

                json = (
                    response if isinstance(response, (dict, list)) else response.json()
                )
                if type(json) == dict:
                    self._set_object_attributes(**json)
            else:
                changed.append(False)
        return changed

    def _alter_properties(self, **properties) -> None:
        """Generic alter method that has to be implemented in child classes
        where arguments will be specified. If a **properties dictionary is empty
        the method will not dispatch any request. If there is at least one entry
        in **properties, an update will be performed according to the update
        type specified in `_API_PATCH` dictionary. A message with update status
        is printed both on failure and success."""
        if not properties:
            if config.verbose:
                logger.info(
                    f"No changes specified for {type(self).__name__} '{self.name}'."
                )
            return None

        changed = self._send_proper_patch_request(properties)

        if config.verbose and all(changed):
            msg = (
                f"{type(self).__name__} '{self.name}' has been modified on the server. "
                "Your changes are saved locally."
            )
            logger.info(msg)

    def _update_nested_properties(
        self,
        objects: Any | list[Any],
        path: str,
        op: str,
        existing_ids: Optional[list[str]] = None,
    ) -> tuple[str, str]:
        """Internal method to update objects with the specified patch wrapper.
        Used for adding and removing objects from nested properties of an
        object like memberships.

        Args:
            objects (Any | list[Any]): An id, object or list of objects
                or/ and object ids on which nested properties should be updated.
            path (str): The name of an attribute that nests other objects.
            op (str): An operation type. Non-ignored values are 'add' or
                'remove'.
            existing_ids (Optional[list[str]], optional): IDs of objects which
                are already nested in `path` attribute. Defaults to None.

        Returns:
            tuple[list[str]]: A tuple whose first element is a list of IDs of
                successful operations, and second element is a list of IDs of
                failed operations.
        """
        from mstrio.access_and_security.privilege import Privilege

        # check whether existing_ids are supplied
        if existing_ids is None:
            existing_ids = [obj.get('id') for obj in getattr(self, path)]

        # create list of objects from strings/objects/lists
        objects_list = objects if isinstance(objects, list) else [objects]
        object_map = {
            obj.id: obj.name for obj in objects_list if isinstance(obj, Entity)
        }

        object_ids_list = [
            obj.id if isinstance(obj, (EntityBase, Privilege)) else str(obj)
            for obj in objects_list
        ]

        # check if objects can be manipulated by comparing to existing values
        if op == "add":
            filtered_object_ids = sorted(
                list(filter(lambda x: x not in existing_ids, object_ids_list))
            )
        elif op == "remove":
            filtered_object_ids = sorted(
                list(filter(lambda x: x in existing_ids, object_ids_list))
            )
        if filtered_object_ids:
            properties = {path: filtered_object_ids}
            self._send_proper_patch_request(properties, op)

        failed = list(sorted(set(object_ids_list) - set(filtered_object_ids)))
        failed_formatted = [
            object_map.get(object_id, object_id) for object_id in failed
        ]
        succeeded_formatted = [
            object_map.get(object_id, object_id) for object_id in filtered_object_ids
        ]
        return (succeeded_formatted, failed_formatted)

    def _validate_type(self, name: str, value: T) -> T:
        """Uses information specified in `self._AVAILABLE_ATTRIBUTES` and
        `self._PATCH_PATH_TYPES` dictionaries to validate type of `value` of an
        attribute with a specified `name`.

        Args:
            name (str): An attribute's name
            value (T): A value to be validated.

        Raises:
            TypeError: If value's type is incorrect.

        Returns:
            T: Unchanged `value` parameter.
        """
        type_map = {**self._AVAILABLE_ATTRIBUTES, **self._PATCH_PATH_TYPES}
        value_type = type_map.get(name, 'Not Found')

        if value_type != 'Not Found' and type(value) != value_type:
            raise TypeError(
                f"'{name}' has incorrect type. Expected type: '{value_type}'"
            )
        return value

    def __setattr__(self, name: str, value: Any) -> None:
        """Overloads the __setattr__ method to validate if this attribute can
        be set for current object. If type of `value` or its literal differs
        from the one that is currently stored, the change will be tracked in a
        dictionary `self._altered_properties` which is later used to update
        properties on a server.
        """

        def track_changes():
            already_tracked = name in self._altered_properties
            if already_tracked:
                current_val = self._altered_properties[name][0]
            elif name in self.__dict__:
                current_val = self.__dict__[name]
            else:
                current_val = None
            is_changed = not (type(current_val) is type(value) and current_val == value)
            if is_changed:
                self._altered_properties.update({name: (current_val, value)})
            elif already_tracked:
                del self._altered_properties[name]

        # Keep track of changed properties if value is already fetched
        if hasattr(self, "_fetched_attributes"):
            if name in self._PATCH_PATH_TYPES and name in self._fetched_attributes:
                track_changes()
            # if value not equal to None then treat as fetched
            if value is not None:
                self._add_to_fetched(name)
        super().__setattr__(name, value)

    def __getattribute__(self, name: str) -> Any:
        """On first fetch, it creates a set `_fetched_attributes` used to track
        attributes that were already fetched from the server. Additionally,
        it checks if there is an API getter in a `_API_GETTERS` dictionary
        that corresponds to the attribute's name. If there is and the object
        has an id and it has not been fetched yet it will fetch the value
        from the server. If not, it will return the value straight away."""
        val = super().__getattribute__(name)

        if name in ["_fetched_attributes", "_find_func"]:
            return val
        if not hasattr(self, "_fetched_attributes"):
            self._fetched_attributes = set()
        if hasattr(self, "_fetched_attributes") and hasattr(self, "_find_func"):
            _name = name[1:] if name.startswith("_") else name
            was_fetched = _name in self._fetched_attributes
            can_fetch = (
                self._find_func(_name) is not None and "id" in self._fetched_attributes
            )
            if can_fetch and not was_fetched:
                self.fetch(_name)  # fetch the relevant object data
            val = super().__getattribute__(name)

        return val

    def _set_dates(self, **kwargs):
        """Transform all date strings provided in kwargs to datetime object.
        Date format strings have to be provided in `_FROM_DICT_MAP`"""
        kwargs = bulk_str_to_datetime(kwargs, self._FROM_DICT_MAP)
        for key, val in self._FROM_DICT_MAP.items():
            if isinstance(val, DatetimeFormats):
                if kwargs.get(key):
                    setattr(self, key, kwargs.get(key))
                elif len(key) > 1 and key[0] == '_' and kwargs.get(key[1:]):
                    setattr(self, key, kwargs.get(key[1:]))

    @property
    def connection(self) -> Connection:
        """An object representation of MicroStrategy connection specific to the
        object."""
        return self._connection

    @property
    def id(self) -> str:
        """The object's id."""
        return self._id

    @property
    def type(self) -> ObjectTypes:
        """The object's type."""
        return self._type


class Entity(EntityBase, ACLMixin, DependenceMixin):
    """Base class representation of the MSTR object.

    Provides methods to fetch, update, and view the object. To implement
    this base class all class attributes have to be provided.

    Attributes:
        connection: A MicroStrategy connection object
        id: Object ID
        name: Object name
        description: Object description
        abbreviation: Object abbreviation
        type: Object type
        subtype: Object subtype
        ext_type: Object extended type
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        version: Version ID
        owner: Owner ID and name
        icon_path: Object icon path
        view_media: View media settings
        ancestors: List of ancestor folders
        certified_info: Certification status, time of certification, and
            information about the certifier (currently only for document and
            report)
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    _API_GETTERS: dict[str | tuple, Callable] = {
        (
            'id',
            'name',
            'description',
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'comments',
            'project_id',
            'hidden',
            'target_info',
        ): objects.get_object_info
    }
    _API_PATCH: dict = {
        ('name', 'description', 'abbreviation'): (objects.update_object, 'partial_put')
    }
    _PATCH_PATH_TYPES = {"name": str, "description": str, "abbreviation": str}
    _FROM_DICT_MAP = {
        **EntityBase._FROM_DICT_MAP,
        'ext_type': ExtendedType,
        'date_created': DatetimeFormats.FULLDATETIME,
        'date_modified': DatetimeFormats.FULLDATETIME,
        'acl': [ACE.from_dict],
        'acg': Rights,
    }

    def _init_variables(self, **kwargs) -> None:
        """Initialize variables given kwargs."""
        from mstrio.users_and_groups.user import User
        from mstrio.utils.certified_info import CertifiedInfo

        super()._init_variables(**kwargs)
        self._date_created = map_str_to_datetime(
            "date_created", kwargs.get("date_created"), self._FROM_DICT_MAP
        )
        self._date_modified = map_str_to_datetime(
            "date_modified", kwargs.get("date_modified"), self._FROM_DICT_MAP
        )
        self.description = kwargs.get("description")
        self.abbreviation = kwargs.get("abbreviation")
        self._subtype = kwargs.get("subtype")
        self._ext_type = (
            ExtendedType(kwargs["ext_type"]) if kwargs.get("ext_type") else None
        )
        self._version = kwargs.get("version")
        self._owner = (
            User.from_dict(
                kwargs.get("owner"),
                self.connection,
            )
            if kwargs.get("owner")
            else None
        )
        self._icon_path = kwargs.get("icon_path")
        self._view_media = kwargs.get("view_media")
        self._ancestors = kwargs.get("ancestors")
        self._certified_info = (
            CertifiedInfo.from_dict(kwargs.get("certified_info"), self.connection)
            if kwargs.get("certified_info")
            else None
        )
        self._hidden = kwargs.get("hidden")
        self._project_id = kwargs.get("project_id")
        self._comments = kwargs.get("comments")
        self._target_info = kwargs.get("target_info")
        self._acg = Rights(kwargs.get("acg")) if kwargs.get("acg") else None
        self._acl = (
            [ACE.from_dict(ac, self._connection) for ac in kwargs.get("acl")]
            if kwargs.get("acl")
            else None
        )

    @property
    def subtype(self) -> ObjectSubTypes:
        return self._subtype

    @property
    def ext_type(self) -> ExtendedType:
        return self._ext_type

    @property
    def date_created(self) -> DatetimeFormats:
        return self._date_created

    @property
    def date_modified(self) -> DatetimeFormats:
        return self._date_modified

    @property
    def version(self) -> str:
        return self._version

    @property
    def owner(self):
        return self._owner

    @property
    def icon_path(self) -> str:
        return self._icon_path

    @property
    def view_media(self) -> int:
        return self._view_media

    @property
    def ancestors(self) -> list[dict]:
        return self._ancestors

    @property
    def certified_info(self):
        return self._certified_info

    @property
    def acg(self) -> Rights:
        return self._acg

    @property
    def acl(self) -> list[ACE]:
        return self._acl

    @property
    def hidden(self) -> bool:
        return self._hidden

    @property
    def project_id(self) -> str:
        return self._project_id

    @property
    def comments(self) -> list[str]:
        return self._comments

    @property
    def target_info(self) -> dict:
        return self._target_info


class CopyMixin:
    """CopyMixin class adds creating copies of objects functionality.

    Currently project objects are not supported. Must be mixedin with
    Entity or its subclasses.
    """

    def create_copy(
        self: Entity,
        name: Optional[str] = None,
        folder_id: Optional[str] = None,
        project: Optional['Project | str'] = None,
    ) -> Any:
        """Create a copy of the object on the I-Server.

        Args:
            name: New name of the object. If None, a default name is generated,
                such as 'Old Name (1)'
            folder_id: ID of the destination folder. If None, the object is
                saved in the same folder as the source object.
            project: By default, the project selected when
                creating Connection object. Override `project` to specify
                project where the current object exists.

        Returns:
                New python object holding the copied object.
        """
        if self._OBJECT_TYPE.value in [32]:
            raise NotSupportedError("Copying of object with type 32 is not supported.")

        from mstrio.server.project import Project

        response = objects.copy_object(
            self.connection,
            id=self.id,
            name=name,
            folder_id=folder_id,
            object_type=self._OBJECT_TYPE.value,
            project_id=project.id if isinstance(project, Project) else project,
        )
        return self.from_dict(source=response.json(), connection=self.connection)


class MoveMixin:
    """MoveMixin class adds moving objects functionality.
    Must be mixedin with Entity or its subclasses.
    """

    def move(self: Entity, folder: 'Folder | str'):
        """Move the object to a folder on the I-Server.

        Args:
            folder: Destination folder, specified either by id
                or the Folder object.
        """
        from mstrio.object_management.folder import Folder

        folder = folder.id if isinstance(folder, Folder) else folder
        self._alter_properties(folder_id=folder)

    def _find_object_with_name(
        self,
        connection: Connection,
        name: Optional[str] = None,
        listing_function: callable = None,
    ) -> dict:
        """Find objects with given name if no id is given.

        Args:
            connection: A MicroStrategy connection object
            name: name of the object. Defaults to None.
            object_type: type of an object that is searched.
            listing_function: function called to list all the objects
                with given name

        Returns:
            dict: object properties in a dictionary.

        Raises:
            ValueError: if both `id` and `name` are not provided,
                if there is more than 1 object with the given `name` or
                if object with the given `name` doesn't exist.
        """
        if name is None:
            raise ValueError(
                "Please specify either 'name' or 'id' parameter in the constructor."
            )
        objects = listing_function(connection=connection, name=name)
        if objects:
            number_of_objects = len(objects)
            if 1 < number_of_objects <= 5:
                error_string = (
                    f"There are {number_of_objects} {self.__class__.__name__} objects"
                    " with this name. Please initialize with id.\n"
                )
                for object in objects:
                    error_string += 'Folder path: '
                    for d in object.ancestors:
                        error_string += d['name'] + '/'
                    error_string += (
                        f' {self.__class__.__name__} id: {object.id} - '
                        f'{self.__class__.__name__} name: {object.name}\n'
                    )
                raise ValueError(error_string)
            if number_of_objects > 5:
                raise ValueError(
                    f"There are {number_of_objects} {self.__class__.__name__}"
                    " objects with this name. Please initialize with id."
                )
            return objects[0].to_dict()
        else:
            raise ValueError(
                f"There is no {self.__class__.__name__} with the given name: '{name}'"
            )


class DeleteMixin:
    """DeleteMixin class adds deleting objects functionality.
    Must be mixedin with Entity or its subclasses.
    """

    _delete_confirm_msg: Optional[str] = None
    _delete_success_msg: Optional[str] = None

    def delete(self: Entity, force: bool = False) -> bool:
        """Delete object.

        Args:
            force: If True, then no additional prompt will be shown before
                deleting object.

        Returns:
            True on success. False otherwise.
        """
        object_name = self.__class__.__name__

        user_input = 'N'
        if not force:
            message = self._delete_confirm_msg or (
                f"Are you sure you want to delete {object_name} "
                f"'{self.name}' with ID: {self._id}? [Y/N]: "
            )
            user_input = input(message) or 'N'
        if force or user_input == 'Y':
            param_value_dict = auto_match_args_entity(self._API_DELETE, self)
            response = self._API_DELETE(**param_value_dict)
            if response.status_code == 204 and config.verbose:
                msg = (
                    self._delete_success_msg
                    or f"Successfully deleted {object_name} with ID: '{self._id}'."
                )
                logger.info(msg)
            return response.ok
        else:
            return False


class CertifyMixin:
    """CertifyMixin class adds certifying and decertifying functionality.
    Must be mixedin with Entity or its subclasses.
    """

    def _toggle_certification(
        self: Entity, certify: bool = True, success_msg: str = None
    ) -> bool:
        object_name = self.__class__.__name__
        expected_result = 'certified' if certify else 'decertified'

        self.fetch()
        if certify == self.certified_info.certified:
            logger.warning(
                f"The {object_name} with ID: '{self._id}' is already {expected_result}"
            )
            return True

        response = objects.toggle_certification(
            connection=self._connection,
            id=self._id,
            object_type=self._OBJECT_TYPE.value,
            certify=certify,
        )
        if response.ok and config.verbose:
            self._set_object_attributes(**response.json())
            id_ = self._id
            msg = (
                success_msg
                or f"The {object_name} with ID: '{id_}' has been {expected_result}."
            )
            logger.info(msg)
        return response.ok

    def certify(self: Entity) -> bool:
        """Certify object.
        Args:
           success_msg: Custom message displayed on success.
        Returns:
                True on success, False otherwise."""
        return self._toggle_certification(certify=True)

    def decertify(self: Entity) -> bool:
        """Decertify object.
        Args:
           success_msg: Custom message displayed on success.
        Returns:
                True on success, False otherwise."""
        return self._toggle_certification(certify=False)


class VldbMixin:
    """VLDBMixin class adds vldb management for supporting objects.

    Objects currently supporting VLDB settings are dataset, document, dossier.
    Must be mixedin with Entity or its subclasses.
    """

    _parameter_error = "Please specify the project parameter."

    def list_vldb_settings(self: Entity, project: Optional[str] = None) -> list:
        """List VLDB settings."""
        connection = (
            self.connection if hasattr(self, 'connection') else self._connection
        )
        if not project and not connection.project_id:
            raise ValueError(self._parameter_error)

        response = objects.get_vldb_settings(
            connection, self.id, self._OBJECT_TYPE.value, project
        )
        return response.json()

    def alter_vldb_settings(
        self: Entity,
        property_set_name: str,
        name: str,
        value: dict,
        project: Optional[str] = None,
    ) -> None:
        """Alter VLDB settings for a given property set."""

        connection = (
            self.connection if hasattr(self, 'connection') else self._connection
        )
        if not project and not connection.project_id:
            raise ValueError(self._parameter_error)

        body = [{"name": name, "value": value}]
        response = objects.set_vldb_settings(
            connection,
            self.id,
            self._OBJECT_TYPE.value,
            property_set_name,
            body,
            project,
        )
        if config.verbose and response.ok:
            logger.info('VLDB settings altered.')

    def reset_vldb_settings(self: Entity, project: Optional[str] = None) -> None:
        """Reset VLDB settings to default values."""

        connection = (
            self.connection if hasattr(self, 'connection') else self._connection
        )
        if not project and not connection.project_id:
            raise ValueError(self._parameter_error)

        response = objects.delete_vldb_settings(
            connection, self.id, self._OBJECT_TYPE.value, project
        )
        if config.verbose and response.ok:
            logger.info('VLDB settings reset to default.')


def auto_match_args_entity(
    func: Callable, obj: EntityBase, exclude: list = [], include_defaults: bool = True
) -> dict:
    """Automatically match `obj` object data to function arguments.

    Handles default parameters. Extracts value from Enums. Returns matched
    arguments as dict.

    Args:
        function: function for which args will be matched
        obj: object to use for matching the function args
        exclude: set `exclude` parameter to exclude specific param-value pairs
        include_defaults: if `False` then values which have the same value as
            default will not be included in the result
    Raises:
        KeyError: could not match all required arguments
    """
    # convert names starting with '_'
    obj_dict = {
        key[1:] if key.startswith("_") else key: val
        for key, val in obj.__dict__.items()
    }
    kwargs = helper.auto_match_args(func, obj_dict, exclude, include_defaults)

    if "object_type" in kwargs:
        kwargs.update({"object_type": obj._OBJECT_TYPE.value})

    return kwargs



from typing import Optional

from mstrio.connection import Connection
from mstrio.distribution_services.schedule import Schedule
from mstrio.distribution_services.subscription import Content, Delivery, Subscription
from mstrio.utils.version_helper import class_version_handler


@class_version_handler('11.3.0600')
class HistoryListSubscription(Subscription):
    """Class representation of MicroStrategy History List Subscription
    object."""

    def __init__(
        self,
        connection: Connection,
        id: Optional[str] = None,
        subscription_id: Optional[str] = None,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ):
        """Initializes HistoryListSubscription object and populates it with
        I-Server data if id or subscription_id is passed.
        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is omitted.
        Args:
            connection (Connection): MicroStrategy connection object returned
                by `connection.Connection()`
            id (str, optional): ID of the subscription to be initialized, only
                id or subscription_id have to be provided at once, if both are
                provided id will take precedence
            subscription_id (str, optional): ID of the subscription to be
                initialized
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
        """
        super().__init__(connection, id, subscription_id, project_id, project_name)

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        multiple_contents: Optional[bool] = False,
        allow_delivery_changes: Optional[bool] = None,
        allow_personalization_changes: Optional[bool] = None,
        allow_unsubscribe: Optional[bool] = None,
        send_now: Optional[bool] = None,
        owner_id: Optional[str] = None,
        schedules: Optional[str | list[str] | Schedule | list[Schedule]] = None,
        contents: Optional[Content] = None,
        recipients: Optional[list[dict] | list[str]] = None,
        delivery: Optional[Delivery | dict] = None,
        delivery_expiration_date: Optional[str] = None,
        delivery_expiration_timezone: Optional[str] = None,
        contact_security: bool = True,
        do_not_create_update_caches: bool = True,
        overwrite_older_version: bool = False,
        re_run_hl: bool = True,
        delivery_notification_enabled: bool = False,
        delivery_personal_notification_address_id: Optional[str] = None,
    ) -> "HistoryListSubscription":
        """Creates a new history list subscription.
        Args:
            connection (Connection): a MicroStrategy connection object
            name (str): name of the subscription
            project_id (str, optional): project ID
            project_name (str, optional): project name
            multiple_contents (bool, optional): whether multiple contents are
                allowed
            allow_delivery_changes (bool, optional): whether the
                recipients can change the delivery of the subscription
            allow_personalization_changes (bool, optional): whether
                the recipients can personalize the subscription
            allow_unsubscribe (bool, optional): whether the recipients
                can unsubscribe from the subscription
            send_now (bool, optional): indicates whether to execute the
                subscription immediately
            owner_id (str, optional): ID of the subscription owner, by
                default logged in user ID
            schedules (list[str] | str | Schedule | list[Schedule], optional):
                Schedules IDs or Schedule objects
            contents (Content, optional): The content settings
            recipients (list[dict] | dict, optional): list of recipients IDs
                or dicts
            delivery (dict | Delivery, optional): delivery object or dict
            delivery_expiration_date (str, optional): expiration date
                of the subscription, format should be yyyy - MM - dd
            delivery_expiration_timezone (str, optional): expiration timezone
                of the subscription, example value 'Europe/London'
            contact_security (bool): whether to use contact security for each
                contact group member
            do_not_create_update_caches (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                cache in the history list
            overwrite_older_version (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list
            re_run_hl (bool): whether subscription will re-run against warehouse
            delivery_notification_enabled (bool): Whether notification is
                enabled, notification applies to cache
            delivery_personal_notification_address_id (str, optional):
                Notification details
        """
        notification_address_id = delivery_personal_notification_address_id

        return super()._Subscription__create(
            connection=connection,
            name=name,
            project_id=project_id,
            project_name=project_name,
            multiple_contents=multiple_contents,
            allow_delivery_changes=allow_delivery_changes,
            allow_personalization_changes=allow_personalization_changes,
            allow_unsubscribe=allow_unsubscribe,
            send_now=send_now,
            owner_id=owner_id,
            schedules=schedules,
            contents=contents,
            recipients=recipients,
            delivery=delivery,
            delivery_mode=Delivery.DeliveryMode.HISTORY_LIST,
            delivery_expiration_date=delivery_expiration_date,
            delivery_expiration_timezone=delivery_expiration_timezone,
            contact_security=contact_security,
            do_not_create_update_caches=do_not_create_update_caches,
            overwrite_older_version=overwrite_older_version,
            re_run_hl=re_run_hl,
            delivery_notification_enabled=delivery_notification_enabled,
            delivery_personal_notification_address_id=notification_address_id,
        )

    def alter(
        self,
        name: Optional[str] = None,
        multiple_contents: Optional[bool] = None,
        allow_delivery_changes: Optional[bool] = None,
        allow_personalization_changes: Optional[bool] = None,
        allow_unsubscribe: Optional[bool] = None,
        send_now: Optional[bool] = None,
        owner_id: Optional[str] = None,
        schedules: Optional[str | list[str] | Schedule | list[Schedule]] = None,
        recipients: Optional[list[dict] | list[str]] = None,
        delivery: Optional[Delivery | dict] = None,
        custom_msg: Optional[str] = None,
        delivery_expiration_date: Optional[str] = None,
        delivery_expiration_timezone: Optional[str] = None,
        contact_security: bool = True,
        do_not_create_update_caches: bool = True,
        overwrite_older_version: bool = False,
        re_run_hl: bool = True,
        delivery_notification_enabled: bool = False,
        delivery_personal_notification_address_id: Optional[str] = None,
    ):
        """Alter the subscription.
        Args:
            name (str, optional): name of the subscription
            multiple_contents (bool, optional): whether multiple contents are
                allowed
            allow_delivery_changes (bool, optional): whether the
                recipients can change the delivery of the subscription
            allow_personalization_changes (bool, optional): whether
                the recipients can personalize the subscription
            allow_unsubscribe (bool, optional): whether the recipients
                can unsubscribe from the subscription
            send_now (bool, optional): indicates whether to execute the
                subscription immediately
            owner_id (str, optional): ID of the subscription owner, by
                default logged in user ID
            schedules (list[str] | str | Schedule | list[Schedule], optional):
                Schedules IDs or Schedule objects
            recipients (list[dict] | dict, optional): list of recipients IDs
                or dicts
            delivery (dict | Delivery, optional): delivery object or dict
            delivery_expiration_date (str, optional): expiration date
                of the subscription, format should be yyyy - MM - dd
            delivery_expiration_timezone (str, optional): expiration timezone
                of the subscription
            contact_security (bool): whether to use contact security for each
                contact group member
            do_not_create_update_caches (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                cache in the history list
            overwrite_older_version (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list
            re_run_hl (bool): whether subscription will re-run against
                warehouse
            delivery_notification_enabled (bool): Whether notification is
                enabled, notification applies to cache
            delivery_personal_notification_address_id (str, optional):
                Notification details
        """
        notification_address_id = delivery_personal_notification_address_id

        return super().alter(
            name=name,
            multiple_contents=multiple_contents,
            allow_delivery_changes=allow_delivery_changes,
            allow_personalization_changes=allow_personalization_changes,
            allow_unsubscribe=allow_unsubscribe,
            send_now=send_now,
            owner_id=owner_id,
            schedules=schedules,
            recipients=recipients,
            delivery=delivery,
            custom_msg=custom_msg,
            delivery_expiration_date=delivery_expiration_date,
            delivery_expiration_timezone=delivery_expiration_timezone,
            contact_security=contact_security,
            do_not_create_update_caches=do_not_create_update_caches,
            overwrite_older_version=overwrite_older_version,
            re_run_hl=re_run_hl,
            delivery_notification_enabled=delivery_notification_enabled,
            delivery_personal_notification_address_id=notification_address_id,
        )



"""Remove each address from every departed user (those which are disabled).
For each successfully removed address a message will be printed.
"""

from typing import List

from mstrio.connection import Connection, get_connection
from mstrio.users_and_groups import list_users


def delete_addresses_from_departed_users(connection: "Connection") -> List[dict]:
    """Remove each address from every departed user (those which are disabled).
    For each successfully removed address a message will be printed.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
    """

    # get all users that are disabled
    all_users = list_users(connection=connection)
    users_ = [u for u in all_users if not u.enabled]
    removed_addresses = []
    for user_ in users_:
        # remove all email addresses from the given user
        if user_.addresses:
            for address in user_.addresses:
                user_.remove_address(id=address['id'])

                removed_addresses.append(
                    {
                        'user_id': user_.id,
                        'username': user_.username,
                        'address': address,
                    }
                )
    return removed_addresses


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# execute deletion of addresses from departed users
delete_addresses_from_departed_users(conn)
